#!/usr/bin/env python
#coding=utf-8

from captureAgents import AgentFactory
from basic_agents import BasicAgent
import random
from game import Directions

''' These are some really simple mice that do not know very much about the maze.
    They don't compute and shortest paths but instead use a set of simple
    heuristics to navigate. Also they only care about food and will not run
    away from or pursue the enemy.

    Random Mouse:
        This is a bit better than a brownian agent. Instead of always making
        random moves it only makes random moves at junctions. In corridors, it
        keeps going forward. If it hits a dead-end it will turn around.

    Danger Mouse:
        This is a bit better than the RagndomMouse. This mouse knows about the
        cheese. Instead of making random moves at junctions it trys to move
        towards its current food. The current food is selected at random in the
        beginning, and then then it selects the closest. It always uses
        manhatten distance. If it can't move towards its current food at the
        given junction, it will make right hand turns until it can. If it gets
        stuck in a cycle (keeps moving through the same number of positions)
        then it will try to make a few right had turns until it can move towards
        the food again.

    The agents serve to demonstrate some basic techniques such storing game
    state position and moves, and how to navigate without knowing anything about
    the maze other than what can be seen. In pacman however you know what the
    maze looks like, and you should probably use a shortest-path algorithm to
    navigate from A to B. This will most certainly be more effective than these
    poor mice :-)

'''

def manhatten_distance(pos1, pos2):
    return abs(pos1[0]-pos2[0])+ abs(pos1[1]-pos2[1])

def direction_tuple(pos1, pos2):
        ''' direction from pos1 to pos2 as tuple of game.Directions'''
        results = []
        xdiff = pos1[0] - pos2[0]
        ydiff = pos1[1] - pos2[1]
        if xdiff < 0:
            results.append(Directions.EAST)
        elif xdiff > 0:
            results.append(Directions.WEST)
        if ydiff < 0:
            results.append(Directions.NORTH)
        elif ydiff > 0:
            results.append(Directions.SOUTH)
        return results

def detect_cycle(length, positions):
    ''' detect cycles in a position list '''
    return positions[-length:] == positions[-2*length:-length]

def detect_cycles(min_length, max_length, positions):
    for length in range(min_length,max_length+1):
        if detect_cycle(length,positions):
            return True

class MouseAgents(AgentFactory):

    def __init__(self, **args):
        AgentFactory.__init__(self, **args)
        self.danger = True

    def getAgent(self, index):
        #if self.danger:
        #    self.danger = False
        return DangerMouse(index)
        return RandomMouse(index)

class RandomMouse(BasicAgent):

    def register_initial_state(self, game_state):
        # store agent positions and actios, and game state.
        self.position_stack = []
        self.action_stack = []
        self.game_state_stack = []
        self.debug = False
        BasicAgent.register_initial_state(self,game_state)
        print game_state.getWalls()

    @property
    def previous_action(self):
        ''' previous action '''
        # this is a total hack to bootstrap the action stack.
        if len(self.action_stack) is 0:
            return self.get_random_action()
        else:
            return self.action_stack[-1]

    @property
    def game_state(self):
        ''' current game state '''
        return self.game_state_stack[-1]

    @property
    def legal_actions(self):
        ''' current legal actions '''
        return self.game_state.getLegalActions(self.index)

    @property
    def position(self):
        ''' current position '''
        return self.game_state.getAgentPosition(self.index)

    @property
    def food_list(self):
        ''' a somewhat accurate list of the food we need to eat as a list of
        tuples '''
        flist = self.get_food(self.game_state).asList()
        # now throw out all the bogus ones
        # FIXME find a neater way to do this
        result = []
        for truth,food in [(self.game_state.hasFood(f[0], f[1]),f) for f in flist]:
            if truth:
                result.append(food)
        return result

    def choose_action(self, game_state):
        ''' store some information about the game, and choose the next action
        '''
        # store game_state
        self.game_state_stack.append(game_state)
        # store current position
        self.position_stack.append(self.position)
        # get next action, store it, and return it
        action = self.mouse_action()
        self.action_stack.append(action)
        return action

    def get_random_action(self):
        ''' really choose a totally random action except STOP'''
        actions = self.legal_actions
        return random.choice(actions)

    def basic_maze_action(self):
        ''' Defines the basic moves a mouse makes in the maze

            If we are at a junction, execute junction_action
            
            If we are in a corridor, keep going straight

            If we are in a dead-end, turn around and go back out

        '''
        actions = self.legal_actions
        # IMPORTANT = STOP is also an action
        # if STOP is our only action, do it
        if len(actions) == 1:
            return actions[0]
        # now remove STOP from the list of possible actions
        actions.remove(Directions.STOP)

        if len(actions) > 2:
            return self.junction_action()
        if len(actions) == 2:
            # keep going whAere the maze leads you
            return actions[0] if actions[0] is not Directions.REVERSE[self.previous_action] \
            else  actions[1]
        if len(actions) == 1:
            # turn around and come back
            return actions[0]

    def mouse_action(self):
        ''' this is what will be called and cached by chooseAction '''
        return self.basic_maze_action()

    def junction_action(self):
        ''' Basic junction action:  make a random action,
            but don't go back where we came from and don't stop
        '''
        actions = self.legal_actions
        actions.remove(Directions.STOP)
        actions.remove(Directions.REVERSE[self.previous_action])
        return random.choice(actions)

    def get_opponent_indices(self):
        # FIXME this should go into BasicAgent
        if self.is_red:
            return self.game_state.getBlueTeamIndices()
        else:
            return self.game_state.getRedTeamIndices()

class DangerMouse(RandomMouse):

    def register_initial_state(self, game_state):
        RandomMouse.register_initial_state(self, game_state)
        # minimum length of cycles to detec
        self.min_cycle_length = 3
        # maximum size of cycles to detect
        self.max_cycle_length = 6
        # at the beginning we have no idea what food we will want to eat
        self.current_food = None
        # number of evasive turns to make if stuck in cycle
        self.evasive_turns = 6
        self.current_evasive_turns = 0
        # what it is our initial junction behaviour
        # this will be monkey patched at runtime
        self.junction_action = self.move_closer_to_food

    def mouse_action(self):
        if self.current_food is None:
            self.current_food = self.select_random_food()

        # FIXME this is a total hack, since i couldn't figure out
        # when the food had been eaten
        if self.position is self.current_food or \
            self.current_food not in self.food_list or \
            not self.game_state.hasFood(self.current_food[0], self.current_food[1]):
            self.current_food = self.select_closest_food()

        # if we are done right hand turning move closer to food again
        if self.current_evasive_turns is 0:
            self.current_junction_action = self.move_closer_to_food
        else:
            self.current_evasive_turns -= 1

        return self.basic_maze_action()

    def move_closer_to_food(self):
        ''' this selects a direction that would potentially bring us closer to
        food '''

        if detect_cycles(self.min_cycle_length, self.max_cycle_length, self.position_stack):
            self.current_evasive_turns = self.evasive_turns
            self.current_junction_action = self.right_hand_turn
            return self.right_hand_turn()

        my_pos = self.game_state.getAgentPosition(self.index)
        good_directions = direction_tuple(my_pos, self.current_food)

        actions= self.legal_actions
        # if any of the best directions can be done, do them
        for d in good_directions:
            if d in actions and d is not Directions.REVERSE[self.previous_action]:
                return d
        # otherwise try to make a right hand turn to circumvent the obstacle
        return self.right_hand_turn()

    def right_hand_turn(self):
        ''' assuming we are at a junction try to make the most right handed turn
        '''
        # find the next possible right hand turn
        # there has got to be an easy way to do this
        actions = self.legal_actions
        next_action = self.previous_action
        for i in range(4):
            # first check if we can keep going along the wall
            if next_action in actions and next_action is not Directions.REVERSE[self.previous_action]:
                return next_action
            # then try to make the right hand turn
            next_action = Directions.RIGHT[next_action]

    def select_random_food(self):
        ''' from all possible food, select a random one '''
        return random.choice(self.food_list)

    def select_closest_food(self):
        ''' return closest food based on manhatten distance '''
        distances = [manhatten_distance(self.position, f) for f in self.food_list]
        return self.food_list[distances.index(min(distances))]

