import random
from basic_agents import BasicAgent, BasicAgentFactory
from game import Directions
from graph import Graph, Node, Edge, NoPathGraphException
import numpy

class EmergencyAgent(BasicAgent):

    def register_initial_state(self, game_state):
        """This method handles the initial setup of the agent.
        """
        super(EmergencyAgent, self).register_initial_state(game_state)
        self.maze_graph, self.maze_nodes = self.create_maze_graph(game_state)
        self.foods_graph, self.foods_nodes, self.food_paths = self.create_food_graph(game_state)
        self.path_to_follow = []
   
    def choose_action(self, game_state):
        
        agent_position = game_state.getAgentPosition(self.index)
        
        if not self.path_to_follow or \
           agent_position == game_state.getInitialAgentPosition(self.index) or \
           not self.get_food(game_state)[self.path_to_follow[-1][0]][self.path_to_follow[-1][1]] or \
           self.is_there_enemy(game_state):
        
            foods_bool = self.get_food(game_state)
            foods = []
            for i in range(foods_bool.width):
                for j in range(foods_bool.height):
                    if foods_bool[i][j]:
                        foods.append((i,j))
            food_distances = [len(self.get_path(agent_position, food)) for food in foods]
            nearest_food = foods[numpy.argmin(food_distances)]
            
            if random.random() > 0.85:
                food_position = random.choice(foods)
            else:
                food_position = nearest_food
            self.path_to_follow = self.get_path(agent_position, food_position)
        
        if self.path_to_follow:    
            destination = self.path_to_follow[0]
            del self.path_to_follow[0]
            
            action = self.convert_next_position_to_direction(agent_position, destination)
        else:
            action = Directions.STOP
        
        actions = game_state.getLegalActions(self.index)
        if action in actions:
            return action
        else:
            return Directions.STOP
    
    def is_there_enemy(self, game_state):
        
        enemies_pos = self.get_team_positions(game_state, enemy_pos=True)
        for enemy_pos in enemies_pos:
            if enemy_pos: 
                return True
        return False
   
    def manhatten_distance(self, pos1, pos2):
        '''Computes the Manhatten distance'''
        return abs(pos1[0]-pos2[0])+ abs(pos1[1]-pos2[1])
       
    def convert_next_position_to_direction(self, current_pos, next_pos):
        '''Converts a move given with 2 (neighbour) position to a Direction'''
        if current_pos[0] - next_pos[0] == 1:
            return Directions.WEST
        elif current_pos[0] - next_pos[0] == -1:
            return Directions.EAST
        elif current_pos[1] - next_pos[1] == 1:
            return Directions.SOUTH
        elif current_pos[1] - next_pos[1] == -1:
            return Directions.NORTH
        else:
            return Directions.STOP 
    
    def create_maze_graph(self, game_state):
        """Create a graph representation of the non-wall positions"""
        walls = game_state.getWalls()
        width, height = walls.width, walls.height
        
        nodes = {}
        for i in range(width):
            for j in range(height):
                if not walls[i][j]:
                    nodes[(i,j)] = Node((i,j))
        
        edges = []
        for (i,j), node in nodes.iteritems():
            if (i+1,j) in nodes:
                edges.append(Edge(nodes[(i,j)], nodes[(i+1,j)]))
                edges.append(Edge(nodes[(i+1,j)], nodes[(i,j)]))
            if (i,j+1) in nodes:
                edges.append(Edge(nodes[(i,j)], nodes[(i,j+1)]))
                edges.append(Edge(nodes[(i,j+1)], nodes[(i,j)]))
                
        return Graph(edges), nodes
    
    def create_food_graph(self, game_state):
        """Create a graph composed of only enemy foods which we are supposed to eat
        
        The nodes in the graph will have edges to all accessible other nodes"""
        
        foods = self.get_food(game_state)
        
        nodes = {}
        for i in range(foods.width):
            for j in range(foods.height):
                if foods[i][j]:
                    nodes[(i,j)] = Node((i,j))
                  
        edges = []
        food_pairs = [(food1, food2) for (i,food1) in enumerate(nodes) 
                                     for (j,food2) in enumerate(nodes)
                                     if i < j]
        food_paths = {}
        for food1, food2 in food_pairs:
            path1to2 = self.shortest_path(food1,food2)
            path2to1 = [food1] + path1to2[:-1]
            path2to1.reverse()
            edges.append(Edge(nodes[food1], nodes[food2]))
            edges.append(Edge(nodes[food2], nodes[food1]))
            food_paths[(food1, food2)] = path1to2
            food_paths[(food2, food1)] = path2to1
        
        return Graph(edges), nodes, food_paths
    
    def get_path(self, src, dest):
        if src == dest:
            return []
        elif (src, dest) in self.food_paths:
            return self.food_paths[(src, dest)]
        else:
            assert src in self.maze_nodes, "Position %s is not valid" % str(src)
            assert dest in self.maze_nodes, "Position %s is not valid" % str(dest)
            return self.shortest_path(src, dest)
    
    def shortest_path(self, src, dest):
        """Returns the path with the shortest length from src to dest.
        
        src and dest are tuples indicating cell locations. path is a list of tuples.
        Ex: shortest_path((1,2), (2,4)) may return a path which is [(1,3), (2,3), (2,4)]
        It includes the destination but excludes the source
        """
        edges_on_path, path_weight = self.maze_graph.shortest_path(self.maze_nodes[src], self.maze_nodes[dest])
        path = [edge.head.position for edge in edges_on_path]
        return path


class EmergencyAgentFactory(BasicAgentFactory):
    def get_agents_list(self):
        return [EmergencyAgent, EmergencyAgent]
 
