from game import GameStateData
from game import Game, Grid
from nikoGraph import Graph, Node, Edge, NoPathGraphException
import numpy
from multiprocessing import Process, Value, Array
import pickle
import os
import mazeLib

class DistanceCalculation(object):

    def __init__(self):
        self.wall_lookup_table = None
        self.wall_graph = None
        self.already_built_lookup_table = False
        self.already_built_wall_graph = False

    def _compute_maze_distance(self,walls,startpoint,endpoint):
        self._assert_valid_position(walls,startpoint)
        self._assert_valid_position(walls,endpoint)
        self._create_wall_lookup_table(walls)
        return self.wall_lookup_table[startpoint[0]][startpoint[1]][endpoint[0]][endpoint[1]]


    def _compute_maze_path(self,walls,startpoint,endpoint):
        self._assert_valid_position(walls,startpoint)
        self._assert_valid_position(walls,endpoint)
        self._create_wall_lookup_table(walls)
        path,dist = self._shortest_path(walls, startpoint, endpoint)
        return path

    def _create_wall_graph(self,grid):
        if self.wall_graph is not None:
            return
        assert self.already_built_wall_graph == False
        #if (startpoint > grid.height
        self.nodes = [[Node() for _ in range(grid.height)] for _ in range(grid.width)]

        self.edges = []
        for x in range(grid.width):
            for y in range(grid.height):
                if grid[x][y]==True:
                    continue
                if (x>0) and not (grid[x-1][y]):
                    self.edges.append(Edge(self.nodes[x][y],self.nodes[x-1][y]))
                if (x<grid.width-1) and not (grid[x+1][y]):
                    self.edges.append(Edge(self.nodes[x][y],self.nodes[x+1][y]))
                if (y>0) and not (grid[x][y-1]):
                    self.edges.append(Edge(self.nodes[x][y],self.nodes[x][y-1]))
                if (y<(grid.height-1)) and not (grid[x][y+1]):
                    self.edges.append(Edge(self.nodes[x][y],self.nodes[x][y+1]))

        self.wall_graph = Graph(self.edges)
        self.already_built_wall_graph = True


        
    def _create_wall_lookup_table(self,walls):

        if self.wall_lookup_table is not None:
            return
        assert self.already_built_lookup_table == False

        try:
            
            infile = os.path.dirname(mazeLib.__file__) + "/lookuptable_" + str(hash(walls)) + ".lookup"
            ifile=open(infile,'rb')
            self.wall_lookup_table = pickle.load(ifile)
            ifile.close()
            self.already_built_lookup_table = True
            return
        except:
            #print "Lookup Table does not exist, computing ..."
            pass
        
        # build lookup
        self.wall_lookup_table = numpy.ones([walls.width,walls.height,walls.width,walls.height])*numpy.Inf
        
        for x in range(walls.width):
            for y in range(walls.height):
                if walls[x][y]:
                    continue
                for x2 in reversed(range(walls.width)):
                    for y2 in reversed(range(walls.height)):
                        if walls[x][y]:
                            continue
                        if self.wall_lookup_table[x][y][x2][y2] != numpy.Inf:
                            continue
                        if self.wall_lookup_table[x2][y2][x][y] != numpy.Inf:
                            continue
                        try:
                            path,dist = self._shortest_path(walls, (x,y), (x2,y2))
                        except NoPathGraphException:
                            path = None
                            dist = numpy.Inf
                        except:
                            raise
                        self.wall_lookup_table[x][y][x2][y2]=dist
                        self.wall_lookup_table[x2][y2][x][y]=dist

                        # use path to compute more distances
                        if (dist != numpy.Inf) and (len(path) > 1):
                            current_distance = dist
                            for pos in path[1:-1]:
                                assert pos!=(x,y)
                                current_distance -= 1
                                self.wall_lookup_table[pos[0]][pos[1]][x2][y2]=current_distance
                                self.wall_lookup_table[x2][y2][pos[0]][pos[1]]=current_distance
                            
                        
        #print self.wall_lookup_table
        self.already_built_lookup_table = True
        #print "done build looktable"
        #print self.already_built_lookup_table
        
    def _assert_valid_position(self,grid,pos):
        for i in pos:
            assert(i is int(i))
        assert(pos is not None)
        assert(pos[0] in range(grid.width) and pos[1] in range(grid.height))
    
    def _we_are_blue_team(self,gstate,index):
        we_are_blue_team = not (gstate.isOnRedTeam(index))
        #assert(gstate.blueTeam() ^ gstate.redTeam())
        return we_are_blue_team

    def _get_our_food(self,gstate,index):
        we_are_blue_team = self._we_are_blue_team(gstate,index)
        if we_are_blue_team:
            our_food = gstate.getRedFood();
        else:
            our_food = gstate.getBlueFood();
        return our_food
    
    def get_nearest_food_distance(self,gstate,index,position=None):
        """ get_nearest_food_distance(gamestate,index)
        returns the distance,path to nearest food item
        """
        self._assert_valid_position(gstate.getWalls(),position)
        self._create_wall_lookup_table(gstate.getWalls())
        path,dist = self._find_nearest_element(gstate.getWalls(),self._get_our_food(gstate,index),position)
        return dist,path
        
    def get_nearest_enemy_distance(self,gstate,index,position=None):
        self._assert_valid_position(gstate.getWalls(),position)
        self._create_wall_lookup_table(gstate.getWalls())
        self._create_wall_graph(gstate.getWalls())
        we_are_blue_team = self._we_are_blue_team(gstate,index)
        if we_are_blue_team:
            enemy_indices = gstate.getRedTeamIndices()
        else:
            enemy_indices = gstate.getBlueTeamIndices()

        best_distance = numpy.Inf
        agent_distances = gstate.getAgentDistances()
        
        for idx in enemy_indices:
            enemy_position = gstate.getAgentPosition(idx)
            if enemy_position is None:
                current_distance = agent_distances[idx]
            else:
                path,current_distance = self.wall_graph.shortest_path(self.nodes[int(position[0])][int(position[1])],self.nodes[int(enemy_position[0])][int(enemy_position[1])])
            if current_distance < best_distance:
                best_distance = current_distance

        assert best_distance != numpy.Inf
        return best_distance,None

    def get_nearest_food_array(self,gstate,index):
        pass


    def _shortest_path(self, grid, startpoint, endpoint):
        self._assert_valid_position(grid,startpoint)
        self._assert_valid_position(grid,endpoint)
        self._create_wall_graph(grid)
        
        try:
            path, length = self.wall_graph.shortest_path( self.nodes[startpoint[0]][startpoint[1]],self.nodes[endpoint[0]][endpoint[1]])
        except NoPathGraphException:
            path = None
        except:
            raise
            
        # if path is None then there is no path
        if path is None:
            xypath = None
            length = numpy.Inf
        else:
            xypath = self._convert_path2xy(path,self.nodes)
                
        return xypath, length

    def _convert_path2xy(self,path,nodes):
        xypath=[]
        for e in path:
            for y in range(len(nodes)):
                for x in range(len(nodes[y])):
                    if nodes[y][x] is e.tail:
                        xypath.append((y,x))
                        break
    
        for y in range(len(nodes)):
            for x in range(len(nodes[y])):
                if nodes[y][x] is e.head:
                    xypath.append((y,x))
                    break
    
        return xypath


    def _find_nearest_element(self,walls,targetgrid,pos):
        """ find_nearest_element(walls,targetgrid,position)
        input:
                mazegrid from class Grid : walls
                targetgrid from class Grid : desired target points, ie food or enemies
                position [x,y]: starting position
        
        output: best_path,best_distance, where
                best_path [(x,y)(x1,y1)...(xend,yend)]
                best_distance: distance to nearest target

        Example:
                ...
                path,dist = find_nearest_element(walls,food,[10,14])
                if dist<mydist ...
                ...
        """
        self._assert_valid_position(walls,pos)
        self._assert_valid_position(targetgrid,pos)
        self._create_wall_lookup_table(walls)
        best_distance = numpy.Inf
        best_end = None
        for x in range(targetgrid.width):
            for y in range(targetgrid.height):
                if targetgrid[x][y]==False:
                    continue
                # this is lookuptable now
                current_distance = self._compute_maze_distance(walls,pos,(x,y))
                #current_path,current_distance = self._shortest_path(mazegrid,pos,(x,y))
                
                if (current_distance < best_distance):
                    best_endpoint = [x,y]
                    best_distance = current_distance
	#print best_path,best_distance
        best_path = self._compute_maze_path(walls,pos,best_endpoint)
        return best_path,best_distance

if __name__ == '__main__':
    mg = Grid(3,2)
    mg.data = [[1,0],[1,1],[1,1]]
    #print  g
    path,dist = _shortest_path(mg,[0,0],[2,1])
    print path
    fg = Grid(3,2)
    fg.data = [[0,0],[0,1],[0,1]]
    path,dist=_find_nearest_element(mg,fg,[0,0])
    print path
