from game import Agent, Directions
from captureAgents import AgentFactory
import graphicsUtils as gu

class BasicAgentFactory(AgentFactory, object):
    def __init__(self, **args):
        super(BasicAgentFactory, self).__init__(**args)
        self._agents_list = self.get_agents_list()
        self._agent_idx = 0
        
    def get_agents_list(self):
        """Returns a list of Agent subclasses.
        If more agents are requested than provided, return a BasicAgent.
        
        This is the method to override to create new factories."""
        return []
   
    def getAgent(self, index):
        "Returns the agent for the provided index."
        idx = self._agent_idx
        if idx >= len(self._agents_list):
            return BasicAgent(index)        
        agent_class = self._agents_list[self._agent_idx]
        self._agent_idx += 1
        return agent_class(index)

class BasicAgent(Agent, object):
    """A new base class for agents.
    It defines some useful helper methods to interact with the game.
    This class is based on captureAgents.CaptureAgent; I removed stuff
    that I don't think is of general use, and used a more pythonic style.
    
    Interesting internal variables:
        index -- index for this agent
        is_red -- True if you're on the red team, False if you're blue
    """

    def __init__(self, index):
        self.index = index
        self.is_red = None            
        # Access to the graphics (if available)
        self._display = None

    def registerInitialState(self, game_state):
        return self.register_initial_state(game_state)
 
    def register_initial_state(self, game_state):
        """This method handles the initial setup of the agent.
        You should override it if you have any business to do
        before the game begins (e.g., analyzing the maze).
        """
        self.is_red = game_state.isOnRedTeam(self.index)
        
        # Set up access to graphics if available
        import __main__
        if '_display' in dir(__main__):
            self._display = __main__._display
            self._txt_message = ''
            self._txt_graphics = None
            self._txt_pos = None

    # Handle agent's actions
    # ######################

    def getAction(self, game_state):
        """Calls choose_action on a grid position, but continues on
        half positions.
        
        If you subclass BasicAgent, you shouldn't need to override this method.
        """
        self._update_txt_message(game_state)
        
        x, y = game_state.getAgentPosition(self.index)
        if x!=round(x) or y!=round(y):
            # we're halfway from one position to the next
            return game_state.getLegalActions(self.index)[0]
        else:
            return self.choose_action(game_state)

    def choose_action(self, game_state):
        """Return the next action the agent needs to perform.
        This is the method you need to override if you want your
        agent to do something sensible.
        
        Input arguments:
        game_state -- a capture.GameState instance that contain the current
                      state of the game
        """
        return Directions.STOP
        
    # Text messages on screen
    # #######################
        
    def _update_txt_message(self, game_state):
        """Moves the message on screen at the current position
        of the agent."""
        if self._display is not None and 'to_screen' in dir(self._display):
            pos = game_state.getAgentPosition(self.index)
            txt_pos = self._display.to_screen(pos)
            if self.index // 2 == 0:
                anchor = "n"
            else:
                anchor = "s"
                txt_pos = (txt_pos[0], txt_pos[1] - 4)
            if self.index % 2 == 0:
                anchor = anchor + "e"
                txt_pos = (txt_pos[0] - 2, txt_pos[1])
            else:
                anchor = anchor + "w"
                txt_pos = (txt_pos[0] + 2, txt_pos[1])
            if self._txt_graphics:# and (self._txt_pos != txt_pos):
                gu.remove_from_screen(self._txt_graphics)
            self._txt_graphics = gu.text(txt_pos, 'yellow', self._txt_message, 
                                         font='Helvetica', size=9, style='bold',
                                         anchor=anchor)
            self._txt_pos = txt_pos

    def say(self, txt):
        """Display a message on screen."""
        self._txt_message = txt
        self._txt_pos = None
   
    # Helper functions
    # ################
    
    def get_food(self, game_state, enemy_food=True):
        """Return the food on the maze belonging to one of the teams.
        
        Return a game.Grid instance 'food', where food[x][y] == True if
        there is food you should eat or protect (base on the value of
        enemy_food) in that square.

        Keyword arguments:
        enemy_food -- If True, returns the food you're supposed to eat,
                    otherwise return the food you should protect.
                    (default: True)
        """
        if ((self.is_red and enemy_food)
            or (not self.is_red and not enemy_food)):
            return game_state.getBlueFood()
        else:
            return game_state.getRedFood()
        
    def get_team_indices(self, game_state, enemy_indices=True):
        """Return a list of the indices of the agents of one of the teams.
        
        Keyword arguments:
        enemy_indices -- If True, returns the indices of the opposing team,
                    otherwise return the indices of your own team.
                    (default: True)
        """
        if ((self.is_red and enemy_indices)
            or (not self.is_red and not enemy_indices)):
            return game_state.getBlueTeamIndices()
        else:
            return game_state.getRedTeamIndices()
    
    def get_team_positions(self, game_state, enemy_pos=True):
        """Return a list with the position of the members of a team.
        
        If an opponent is too far (> 5 square Manhattan distance),
        the corresponding entry will be None, as its position is unavailable.
        
        Keyword arguments:
        enemy_indices -- If True, returns the positions of the opposing team,
                    otherwise return the position of your own team.
                    (default: True)
        """
        indices = self.get_team_indices(game_state, enemy_indices=enemy_pos)
        positions = [game_state.getAgentPosition(idx) for idx in indices]
        return positions
    
    def get_team_distances(self, game_state, enemy_dist=True):
        """Return a list with the noisy distances from the members of a team.
        Distances will be returned as the real distance, +/- 6.
        
        Keyword arguments:
        enemy_dist -- If True, returns the distance of the opposing team,
                    otherwise return the distance of your own team.
                    (default: True)
        """
        indices = self.get_team_indices(game_state, enemy_indices=enemy_dist)
        distances = game_state.getAgentDistances()
        return [distances[idx] for idx in indices]

    def get_score(self, game_state):
        """Return your score.
        
        Your score is given by a number that is the difference between your
        score and the opponents score. This number is negative if you're losing.
        """
        if self.is_red:
            return game_state.getScore()
        else:
            return -game_state.getScore()

