"""Dijkstra's algorithm for shortest paths
    David Eppstein, UC Irvine, 4 April 2002
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/117228
"""

from priodict import priorityDictionary

def Dijkstra(graph, start, end=None):
    """
    Find shortest paths from the start vertex to all
    vertices nearer than or equal to the end.

    The input graph graph is assumed to have the following
    representation: A vertex can be any object that can
    be used as an index into a dictionary.  graph is a
    dictionary, indexed by vertices.  For any vertex vertx,
    graph[vertx] is itself a dictionary, indexed by the neighbors
    of vertx.  For any edge vertx->wertx, graph[vertx][wertx] is the length of
    the edge.  This is related to the representation in
    <http://www.python.org/doc/essays/graphs.html>
    where Guido van Rossum suggests representing graphs
    as dictionaries mapping vertices to lists of neighbors,
    however dictionaries of edges have many advantages
    over lists: they can store extra information (here,
    the lengths), they support fast existence tests,
    and they allow easy modification of the graph by edge
    insertion and removal.  Such modifications are not
    needed here but are important in other graph algorithms.
    Since dictionaries obey iterator protocol, a graph
    represented as described here could be handed without
    modification to an algorithm using Guido's representation.

    Of course, graph and graph[vertx] need not be Python dict objects;
    they can be any other object that obeys dict protocol,
    for instance a wrapper in which vertices are URLs
    and a call to graph[vertx] loads the web page and finds its links.
    
    The output is a pair (dvar,pvar) where dvar[vertx] is the distance
    from start to vertx and pvar[vertx] is the predecessor of vertx along
    the shortest path from s to vertx.
    
    Dijkstra's algorithm is only guaranteed to work correctly
    when all edge lengths are positive. This code does not
    verify this property for all edges (only the edges seen
     before the end vertex is reached), but will correctly
    compute shortest paths even for some graphs with negative
    edges, and will raise an exception if it discovers that
    a negative edge has caused it to make a mistake.
    """

    dvar = {}    # dictionary of final distances
    pvar = {}    # dictionary of predecessors
    qvar = priorityDictionary()   # est.dist. of non-final vert.
    qvar[start] = 0
    
    for vertx in qvar:
        dvar[vertx] = qvar[vertx]
        if vertx == end: 
            break
        
        for wertx in graph[vertx]:
            vwlength = dvar[vertx] + graph[vertx][wertx]
            if wertx in dvar:
                if vwlength < dvar[wertx]:
                    raise ValueError, \
  "Dijkstra: found better path to already-final vertex"
            elif wertx not in qvar or vwlength < qvar[wertx]:
                qvar[wertx] = vwlength
                pvar[wertx] = vertx
    
    return (dvar, pvar)
            
def shortestPath(graph, start, end):
    """
    Find a single shortest path from the given start vertex
    to the given end vertex.
    The input has the same conventions as Dijkstra().
    The output is a list of the vertices in order along
    the shortest path.
    """

    _, pvar = Dijkstra(graph, start, end)
    path = []
    while 1:
        path.append(end)
        if end == start: 
            break
        end = pvar[end]
    path.reverse()
    return path
