"""This is Group 3's agent for the foraging behavior.
"""

from basic_agents import BasicAgentFactory
from numpy import argsort, zeros
from game import Directions
from path_graph import GraphKnowingAgent
from random import randint, shuffle

class ForagingAgent(GraphKnowingAgent):
    """Group 3's foraging behavior"""
    dir_dict = {(0, 1) : Directions.NORTH, (0, -1) : Directions.SOUTH,
                    (1, 0) : Directions.EAST, (-1, 0) : Directions.WEST, 
                    (0, 0) : Directions.STOP }
    
    def register_initial_state(self, game_state):
        """This method handles the initial setup of the agent.
        It has been overridden since we have business to do
        before the game begins (e.g., analyzing the maze).
        """
        # begin: copied from basic_agent.py
        self.is_red = game_state.isOnRedTeam(self.index)
        super(ForagingAgent, self).register_initial_state(game_state)
        # our code:
        self.construct_graph(game_state)
        self.update_food(game_state)
        randint(0, 5)
        self.infos['pos2'] = (0, 0)
        self.infos['pos1'] = (0, 0)        

    def update_food(self, game_state):
        """Update the nodes with current food information from the game.
        """
        ourfoodgrid = self.get_food(game_state, True)
        theirfoodgrid = self.get_food(game_state, False)
        minority_report = []
        for pos, node in self.nodes.items():
            node.food = int(ourfoodgrid[pos[0]][pos[1]])
            new_their_food = theirfoodgrid[pos[0]][pos[1]]
            node.foodvalue = node.food
            if node.their_food != new_their_food:
                minority_report.append(pos)
            node.their_food = int(new_their_food)
        if len(minority_report) > len(self.get_team_indices(game_state, True)):
            return []
        return minority_report

    def _get_next_food(self, my_pos):
        """Check for food in the immediate vicinity and go to it.
        """
        neighbors = self.get_neighbors(my_pos, 4)
        allit = zeros(len(neighbors.items()))
        tuples = []
        for i in xrange(len(neighbors.items())):     
            allit[i] = (neighbors.items()[i][1]) 
            tuples.append(neighbors.items()[i][0])
        sorted_neighbors = []        
        a_index = argsort(allit)
        for i in a_index:
            sorted_neighbors.append(tuples[i])

        for thy in sorted_neighbors:
            if self.nodes[thy].food:
                self.say('Greedy food')
                return self.get_shortest_path(my_pos, thy)[1]
        return my_pos

    def _flight(self, game_state, my_pos, options):
        """ detect an enemy in the vicinity and choose flight direction"""
        trenches = self.get_team_positions(game_state, enemy_pos=True)
        if trenches[0] == None and trenches[1] == None:
            return my_pos
        
        neighbors = self.get_neighbors(my_pos, 2)
        for thy in neighbors:
            for trench in trenches:
                if thy == trench:
                    bad_path = self.get_shortest_path(my_pos, trench)
                    bad_act = self.get_action(my_pos, bad_path[1], game_state)
                    for action in options:
                        if action != bad_act:
                            return action
        return my_pos
                    
    def _fancy_food(self, my_pos):
        """Fancier food finding algorithm:
            Cannot disclose details due to competing interests.
        """
        foodval_list = []
        pos_list = []
        for pos, node in self.nodes.items():
            for npos, distance in self.get_neighbors(pos, 3).items():
                node.foodvalue += self.nodes[npos].food/(1.0 + distance) ** 2
            foodval_list.append(node.foodvalue)
            pos_list.append(pos)
        indices = argsort(foodval_list)[-5:]
        len_min_path = 999999999
        for idx in indices:
            thispath = self.get_shortest_path(my_pos, pos_list[idx])
            if len(thispath) < len_min_path:
                min_path = thispath
                len_min_path = len(min_path)
        if len(min_path) < 2:
            return min_path[0]
        return min_path[1]

    def _engrave_exp(self, my_pos):
        """Remember the 2 previous moves."""
        self.infos['pos2'] = self.infos['pos1']
        self.infos['pos1'] = my_pos

    def _avoid_friend(self, game_state, my_pos, options):
        """Try to avoid another gatherer."""
        teamindices = self.get_team_indices(game_state, \
            enemy_indices = False)
        teamindices.remove(self.index)
        agent_smith_is = self.infos[teamindices[0]]
        if agent_smith_is == "forager":
            our_pos = self.get_team_positions(game_state, False)
            agent_smith = intify(our_pos[0])
            for pos in our_pos:
                pos = intify(pos)
                if pos != my_pos:
                    agent_smith = pos
                    break
            shuffle(options)
            if agent_smith == my_pos:
                self.say('Get off my property!')
                return options[0]
            bad_path = self.get_shortest_path(my_pos, agent_smith)
            if len(bad_path) < 3: 
                bad_act = self.get_action(my_pos, bad_path[1], game_state)
                for action in options:
                    if action != bad_act:
                        self.say("Let's part ways")
                        return action
        return my_pos

    def choose_action(self, game_state):
        """Returns the action for this Agent"""
        options = ex_stop(game_state.getLegalActions(self.index))
        shuffle(options)
        my_pos = self.get_my_position(game_state)
        self.update_food(game_state)
        rnd = randint(0, len(options) - 1)
        # Stuck in an endless loop?
        if self.infos['pos2'] == my_pos:
            self.say('Trying to break the cycle')
            bad_path = self.get_shortest_path(my_pos, self.infos['pos1'])
            bad_act = self.get_action(my_pos, bad_path[1], game_state)
            for action in options:
                if action != bad_act:
                    return action
        self._engrave_exp(my_pos)
        # Enemy nearby? Run!
        if not self.in_our_half(game_state):
            nextstep = self._flight(game_state, my_pos, options)
            if nextstep != my_pos:
                self.say('Fly, you fools!')
                return nextstep
            # Friend nearby? Run as well!
            nextstep = self._avoid_friend(game_state, my_pos, options)
            if nextstep != my_pos:
                return nextstep
        # Buggy food-eating?
        if self.nodes[my_pos].food:
            self.say('I better wait here')
            return Directions.STOP
        # Food in vicinity?
        nextstep = self._get_next_food(my_pos)
        if nextstep != my_pos:
            return self.get_action(my_pos, nextstep, game_state)
        # Finally, fancy-find the biggest food cache
        nextstep = self._fancy_food(my_pos)
        if nextstep != my_pos:
            self.say('Gathering food')
            return self.get_action(my_pos, nextstep, game_state)
        self.say('No clue what to do!')
        return options[rnd]

    def get_action(self, origin, destination, game_state):
        """Returns the action to be performed to reach the destination 
        from the origin. Origin and Destination have to be 1 apart.
        """
        legals = game_state.getLegalActions(self.index)
        diff = (destination[0] - origin[0], destination[1] - origin[1])
        try:
            action = ForagingAgent.dir_dict[diff]
        except KeyError:
            #print "Origin and destination are not exactly 1 apart."
            raise
        if action in legals:
            return action
        else:
            raise ValueError("No legal move to destination.") 
        
    def get_my_position(self, game_state):
        """Returns the position of this agent as a tuple."""
        our_rank = self.get_team_indices(game_state, False).index(self.index)
        our_position = self.get_team_positions(game_state, False)[our_rank]
        return intify(our_position)
            
    def in_our_half(self, game_state):
        """Returns true if agent is in our part of the maze and hence a 
        ghost."""
        walls = game_state.getWalls()
        width = walls.width   
        our_position = self.get_my_position(game_state)
        
        if self.is_red: # we are red
            return our_position[0] < (1+width)/2 - 1
        else: # we are blue
            return our_position[0] >= (1+width)/2 + 1

class ForagingAgentFactory(BasicAgentFactory):
    """The required Factory for this agent."""
    def get_agents_list(self):
        """Returns two instances of the agent"""
        return [ForagingAgent, ForagingAgent]

def ex_stop(action_list):
    """Removes stop-action from list"""
    if Directions.STOP in action_list:
        action_list.remove(Directions.STOP)
    return action_list

def intify(tuple2):
    """Make the numbers in a 2-tuple into integers."""
    if len(tuple2) == 2:
        return (int(tuple2[0]), int(tuple2[1]))
    else:
        return tuple2
