""""
Group 3
Ghost agent implements strategy for chasing enemies in own half.
"""

import random
from game import Directions
from captureAgents import AgentFactory
import graphicsUtils as gu
import foraging_agents
from foraging_agents import intify

class HuntingAgentFactory(AgentFactory, object):
    """Factory of our HuntingAgent."""
    
    def __init__(self, **args):
        super(HuntingAgentFactory, self).__init__(**args)
        self._agents_list = self.get_agents_list()
        self._agent_idx = 0
        
    def get_agents_list(self):
        """Returns a list of Agent subclasses.
        If more agents are requested than provided, return a GhostAgent.
        
        This is the method to override to create new factories."""
        return []
   
    def getAgent(self, index):
        "Returns the agent for the provided index."
        idx = self._agent_idx
        if idx >= len(self._agents_list):
            return HuntingAgent(index)        
        agent_class = self._agents_list[self._agent_idx]
        self._agent_idx += 1
        return agent_class(index)



class HuntingAgent(foraging_agents.ForagingAgent):
    """A new base class for agents.
    It defines some useful helper methods to interact with the game.
    This class is based on captureAgents.CaptureAgent; I removed stuff
    that I don't think is of general use, and used a more pythonic style.
    
    Interesting internal variables:
        index -- index for this agent
        is_red -- True if you're on the red team, False if you're blue
    """

    def __init__(self, index, infos={}):
        super(HuntingAgent, self).__init__(index, infos)
        self.index = index
        self.is_red = None            
        # Access to the graphics (if available)
        self._display = None
        self._last_move = None 
        self._update_counter = 0
        self._target_position = (None, None)
        self._txt_message = None
        self._txt_graphics = None
        self._txt_pos = None
        random.seed(index)

       
    def manhatten_distance(self, pos1, pos2):
        """Returns well known Manhatten-distance between given positions"""
        return abs(pos1[0]-pos2[0])+ abs(pos1[1]-pos2[1])


    def registerInitialState(self, game_state):
        """Prepare for the game; calculations for path-optimisation; 
        initialising variables depending on game_state.
        """
        self.construct_graph(game_state)
        # make sure not all food is marked as new
        self.assign_new_target_position(game_state)
        return self.register_initial_state(game_state)
 
 
    def exclude_stop_action(self, action_list):
        """Removes stop-action from list"""
        if Directions.STOP in action_list:
            action_list.remove(Directions.STOP)
        return action_list
 
    def register_initial_state(self, game_state):
        """This method handles the initial setup of the agent.
        You should override it if you have any business to do
        before the game begins (e.g., analyzing the maze).
        """
        super(HuntingAgent, self).register_initial_state(game_state)
        self.is_red = game_state.isOnRedTeam(self.index)
        
        # randomly choose a initial movement direction
        actions = game_state.getLegalActions(self.index)
        actions = self.exclude_stop_action(actions)
        self._last_move = random.choice(actions)

        # Set up access to graphics if available
        import __main__
        if '_display' in dir(__main__):
            self._display = __main__._display
            self._txt_message = ''
            self._txt_graphics = None
            self._txt_pos = None

    # Handle agent's actions
    # ######################

    def getAction(self, game_state):
        """Calls choose_action on a grid position, but continues on
        half positions.
        
        If you subclass GhostAgent, you shouldn't need to override this method.
        """
        self._update_txt_message(game_state)
        
        x_pos, y_pos = game_state.getAgentPosition(self.index)
        if x_pos != round(x_pos) or y_pos != round(y_pos):
            # we're halfway from one position to the next
            return game_state.getLegalActions(self.index)[0]
        else:
            return self.choose_action(game_state)


    def get_my_position(self, game_state):
        """Returns the current position of the agent."""
        our_rank = self.get_team_indices(
                game_state, enemy_indices=False).index(self.index)
        our_position = self.get_team_positions(
                game_state, enemy_pos=False)[our_rank]
        return our_position

    def get_current_enemy_positions(self, game_state):
        """Returns list with enemy positions; these are None if enemies are 
        not observable."""
        enemy_positions = self.get_team_positions(
                game_state, enemy_pos=True)
        return enemy_positions
    
    def get_enemy_positions(self, game_state):
        """Returns list with enemy positions; these are None if enemies are 
        not observable. Information is based on shared memory of our agents."""
        # returns enemy positions based on memory    
        enemy_indices = self.get_team_indices(
                        game_state, enemy_indices=True)
        return [ self.retrieve_agent_position(enemy_index) 
            for enemy_index in enemy_indices ]
     
     
    def detect_kill(self, our_position, enemy_position):
        """Returns true if the two given position match with respect to the
        Manhatten distance. Note that this could either mean that we killed 
        an enemy or that we have been eaten ourselves."""
        if self.manhatten_distance( our_position, enemy_position ) == 1.:
            return True
        else:
            return False
        

    def store_agent_position(self, agent_index, position):
        """Adds position of agent to shared dictionary."""
        key = ('position', agent_index)
        self.infos[key] = position
        
    def retrieve_agent_position(self, agent_index):
        """Returns position of agent as stored in shared dictionary if
        available; None otherwise."""
        key = ('position', agent_index)
        try:
            return self.infos[key]
        except KeyError:
            # position has not been stored yet
            return None    

    def update_agent_position_memory(self, game_state):
        """Function to update memory on position information of 
        enemies."""     
        positions = self.get_current_enemy_positions(game_state)
        our_position = self.get_my_position(game_state)
        enemy_indices = self.get_team_indices(
                game_state, enemy_indices=True )
        
        # check if stuck with old invalid memory
        if self.no_move():
            for enemy_index in enemy_indices:  
                self.store_agent_position(enemy_index, None)    
 
        # indirect inference of enemy position through food disappearance
        food_change_list = self.update_food(game_state)
        if len(food_change_list) <= len(enemy_indices):
            # find closest field where food disappears
            min_length = 99999999999999
            closest_position = None
            for position in food_change_list:
                # get shortest path
                path = self.get_shortest_path(our_position, position)
                if len(path) < min_length:
                    min_length = len(path)
                    closest_position = position
                for enemy_index in enemy_indices:
                    # set positions of all agents to determines position
                    
                    previous_position = self.retrieve_agent_position(enemy_index)
                    if previous_position: 
                        if self.detect_kill(our_position, previous_position):  
                            # enemy has been eaten; so reset position to None
                            self.store_agent_position(enemy_index, None)
                        else:
                            self.store_agent_position(
                                    enemy_index, closest_position)
       
 #                   stored_position = self.retrieve_agent_position(enemy_index)
 #                   if not stored_position:
 #                       self.store_agent_position(enemy_index, position)
 #                       break
       
        # check which agents are within visible range
        for position, enemy_index in zip(positions, enemy_indices):         
            if position: # i.e. != None
                # enemy agent not reached yet; store position information
                self.store_agent_position(enemy_index, position)          
            else:
                previous_position = self.retrieve_agent_position(enemy_index)
                if previous_position: 
                    if self.detect_kill(our_position, previous_position):  
                        # enemy has been eaten; so reset position to None
                        self.store_agent_position(enemy_index, None)
                 
    def get_ahdistance_to_frontier(self, game_state):
        """Returns horizontal distance to border as an absolute number 
        (not manhatten distance)."""
        walls = game_state.getWalls()
        width = walls.width   
        our_position = self.get_my_position(game_state)
        return abs( our_position[0] - (1+width)/2 )
        
        
    def sort_actions_by_preference(self, game_state, 
            action_list, preference_function): 
        """Returns a new action list in which actions are sorted according to 
        the score of the preference function"""   
        score_action_list = []
        for action in action_list:
            potential_state = game_state.generateSuccessor(
                    self.index, action)
            score = preference_function(potential_state)
            score_action_list.append( (score, action) )
        score_action_list.sort()
        return [action for score, action in score_action_list]
    
    
    def enemy_visible(self, game_state):
        """Returns true if any enemy is within sight and false otherwise."""
        position_list = self.get_enemy_positions(game_state)
        # see if entry != None exists
        truth = False
        for position in position_list:
            truth = truth or position != None
        return truth
    
    def move_towards_closest_enemy_in_our_half(self, game_state, 
            default_position):
        """Returns action to approach enemy on shortest path; if enemy not in 
        our half go for the one with the longer path; approaching default 
        position otherwise."""
        walls = game_state.getWalls()
        width = walls.width
        # get paths to enemies if multiple observable
        my_pos = self.get_my_position(game_state)
        positions = self.get_enemy_positions(game_state)
        shortest_path = None
        for position in positions:
            min_length = 99999999999999
            if position: # position position != None
                if self.position_in_our_half(width, position):                 
                    # enemy in our half
                    # get shortest path
                    path = self.get_shortest_path(my_pos, position)
                    if len(path) < min_length:
                        min_length = len(path)
                        shortest_path = path
        if not shortest_path:
            # no enemy visible or in memory
            # go to default position
            return self.get_step_towards_target_position(game_state, 
                default_position)
        else: # find step toward enemy
            # identify shortest enemy pass
            # return move
            try:
                first_field_on_shortest_path = shortest_path[1]
            except IndexError:
                # we are already at target position,
                # shortest path thus only contain 1 element 
                return random.choice(self.exclude_stop_action( game_state.getLegalActions(self.index)) )
#                return Directions.STOP
            # move toward enemy
            return self.get_action(my_pos, first_field_on_shortest_path, 
                game_state)
     
    
    def move_towards_closest_enemy(self, game_state, default_position):
        """Returns action to approach enemy on shortest path (if enemey 
        visible); approaching default position otherwise."""
        # get paths to enemies if multiple observable
        my_pos = self.get_my_position(game_state)
        positions = self.get_enemy_positions(game_state)
        shortest_path = None
        for position in positions:
            min_length = 99999999999999
            if position:
                # position position != None
                # get shortest path
                path = self.get_shortest_path(my_pos, position)
                if len(path) < min_length:
                    min_length = len(path)
                    shortest_path = path
        if not shortest_path:
            # no enemy visible or in memory
            # go to default position
            return self.get_step_towards_target_position(game_state, 
                default_position)
        else: # find step toward enemy
            # identify shortest enemy pass
            # return move
            try:
                first_field_on_shortest_path = shortest_path[1]
            except IndexError:
                # we are already at target position,
                # shortest path thus only contain 1 element 
                return Directions.STOP
            # move toward enemy
            return self.get_action(my_pos, first_field_on_shortest_path, 
                game_state)
                 
    
    def get_step_towards_target_position(self, game_state, target_position):
        """Returns movement action on shortest path to target position.
        Action stop is returned if current position equals target."""
        my_pos = self.get_my_position(game_state)
        try:
            first_field_on_shortest_path = self.get_shortest_path(
                    my_pos, target_position)[1]
        except IndexError:
            # we are already at target position,
            # shortest path thus only contain 1 element 
            return Directions.STOP
        return self.get_action(my_pos, first_field_on_shortest_path, game_state)
        
    
    def assign_new_target_position(self, game_state):
        """Updates agent's default position and sets it somewhere close to 
        the border of our half."""
        walls = game_state.getWalls()
        x_pos = walls.width / 2  - 2
        min_y = 4
        max_y = walls.height - 4 - 1
        y_pos = random.randint(min_y, max_y)     
        # set target to non-wall field   
        self._target_position = \
                self.get_valid_positions_as_close_to_target_position(
                game_state, (x_pos, y_pos) )[0]

    def get_closest_safe_position(self, game_state):
        """Returns position close to border in our half with shortest path
        from current position."""
        my_pos = self.get_my_position(game_state)
        walls = game_state.getWalls()
        h_pos = walls.width / 2 - 1
        min_length = 99999999999999
        closest_position = None
        for x_pos, y_pos in walls.asList(False):
            if x_pos != h_pos:
                continue
            else:
                position = (x_pos, y_pos)
                # get shortest path
                path = self.get_shortest_path(my_pos, position)
                if len(path) < min_length:
                    min_length = len(path)
                    closest_position = position
        return closest_position
    
    
    def in_our_half(self, game_state):
        """Returns true if agent is in our part of the maze and hence a 
        ghost."""
        walls = game_state.getWalls()
        width = walls.width   
        our_position = self.get_my_position(game_state)
        return self.position_in_our_half(width, our_position)
  
    def position_in_our_half(self, maze_width, position):
        """Returns true if given position is on our side of the middle of the
        maze."""
        if self.is_red: # we are red
            return position[0] < (1+maze_width)/2 # - 1
        else: # we are blue
            return position[0] >= (1+maze_width)/2 # + 1
        

    def get_valid_positions_as_close_to_target_position(    
            self, game_state, target_position):
        """Returns target position (as list) if not a wall and list of closest 
        (Manhattan distance) surrounding non-wall positions otherwise."""
        walls = game_state.getWalls()
        non_wall_list = []
        for x_pos, y_pos in walls.asList(False):
            # append tupel with (distance, field-coordinates)
            non_wall_list.append( 
    ( self.manhatten_distance( target_position, (x_pos, y_pos) ) , 
            (x_pos, y_pos) ) )
        # sort by distance to target field
        non_wall_list.sort()
        return [ field for dist, field in non_wall_list ]             
        

    def in_enemy_half(self, game_state):
        """Returns true if agent is in enemy half and should be eating."""
        return not self.in_our_half(game_state)

    def no_move(self):
        """Returns true if we did stay on the same field for 2 rounds."""
        try:
            return intify(self.infos['pos1']) == intify(self.infos['pos2'])
        except KeyError:
            return False

    def choose_action(self, game_state):
        """Return the next action the agent needs to perform.
        This is the method you need to override if you want your
        agent to do something sensible.
        
        Input arguments:
        game_state -- a capture.GameState instance that contain the current
                      state of the game
        """
        if self._update_counter % 10 == 0:
            self.say(random.choice([
            'I am a ghost and I know it!', 
            "I'm a defender!", 
            'I am a ghost and proud of it!',
            "Don't get near me",
            "I like killing!",
            "I can smell you",
            "Hasta la vista, Pakki"        
            ]))
        self._update_counter += 1
        my_pos = self.get_my_position(game_state)
        self._engrave_exp(my_pos)
        self.update_agent_position_memory(game_state)
 
# test for get_step_towards_target_position       
#        return self.get_step_towards_target_position(game_state, (5+8,5))
# test for move_towards_closest_enemy
#        return self.move_towards_closest_enemy(game_state, (13,5) )
        return self.strategy_2(game_state)
    
    
    def strategy_2(self, game_state):
        """go to default if no enemy close"""
        
        if self._update_counter % 17 == 0:
            self.assign_new_target_position(game_state)
        default_position = self._target_position
        
        if self.in_our_half(game_state):              
            # get first step towards closest enemy   
#            approach_enemy = self.move_towards_closest_enemy(
#                    game_state, default_position )
            approach_enemy = self.move_towards_closest_enemy_in_our_half( 
                    game_state, default_position )
            # see what happens if we move as planned
            potential_state = game_state.generateSuccessor(
                    self.index, approach_enemy)
            # make sure we are still in our half
            if self.in_our_half(potential_state):
                # not leaving our half when chasing enemy
                return approach_enemy
            else:
                # what to do otherwise?
                return self.strategy_1(game_state)
        else:
            # get safe quickly!
            default_position = \
                    self.get_closest_safe_position(game_state)          
            return self.get_step_towards_target_position(
                game_state, default_position)    
          
            
       
    def strategy_1(self, game_state):
        """Remain in our half of the fiels.
        Keep current heading if possible.
        Minor random variations."""        
        # get moves without crash into wall
        actions = game_state.getLegalActions(self.index)
        # do not consider stop-action
        actions = self.exclude_stop_action(actions)
        # avoid moving out of own field
        safe_actions = []
        for action in actions:
            potential_state = game_state.generateSuccessor(self.index, action)
            if self.in_our_half(potential_state):
                safe_actions.append(action)
        # if current heading not safe choose a new one
        if self._last_move not in safe_actions:
            # make strategic or random decision
            if random.random() < 0.5:
                self._last_move = random.choice(safe_actions)
            else:
                preference_function = self.get_ahdistance_to_frontier
                sorted_actions = self.sort_actions_by_preference(
                        game_state, safe_actions, preference_function) 
                self._last_move = sorted_actions[0]
        else:
            if random.random() < 0.05:
                self._last_move = random.choice(safe_actions)
            else:
                # keep on moving current direction
                pass
     
        # return chosen action
        return self._last_move

 
    # Text messages on screen
    # #######################
        
    def _update_txt_message(self, game_state):
        """Moves the message on screen at the current position
        of the agent."""
        if self._display is not None and 'to_screen' in dir(self._display):
            pos = game_state.getAgentPosition(self.index)
            txt_pos = self._display.to_screen(pos)
            if self.index // 2 == 0:
                anchor = "n"
            else:
                anchor = "s"
                txt_pos = (txt_pos[0], txt_pos[1] - 4)
            if self.index % 2 == 0:
                anchor = anchor + "e"
                txt_pos = (txt_pos[0] - 2, txt_pos[1])
            else:
                anchor = anchor + "w"
                txt_pos = (txt_pos[0] + 2, txt_pos[1])
            if self._txt_graphics:# and (self._txt_pos != txt_pos):
                gu.remove_from_screen(self._txt_graphics)
            self._txt_graphics = gu.text(txt_pos, 'yellow', self._txt_message, 
                                         font='Helvetica', size=9, style='bold',
                                         anchor=anchor)
            self._txt_pos = txt_pos

    def say(self, txt):
        """Display a message on screen."""
        self._txt_message = txt
        self._txt_pos = None
   
    # Helper functions
    # ################
    
    def get_food(self, game_state, enemy_food=True):
        """Return the food on the maze belonging to one of the teams.
        
        Return a game.Grid instance 'food', where food[x][y] == True if
        there is food you should eat or protect (base on the value of
        enemy_food) in that square.

        Keyword arguments:
        enemy_food -- If True, returns the food you're supposed to eat,
                    otherwise return the food you should protect.
                    (default: True)
        """
        if ((self.is_red and enemy_food)
            or (not self.is_red and not enemy_food)):
            return game_state.getBlueFood()
        else:
            return game_state.getRedFood()
        
    def get_team_indices(self, game_state, enemy_indices=True):
        """Return a list of the indices of the agents of one of the teams.
        
        Keyword arguments:
        enemy_indices -- If True, returns the indices of the opposing team,
                    otherwise return the indices of your own team.
                    (default: True)
        """
        if ((self.is_red and enemy_indices)
            or (not self.is_red and not enemy_indices)):
            return game_state.getBlueTeamIndices()
        else:
            return game_state.getRedTeamIndices()
    
    def get_team_positions(self, game_state, enemy_pos=True):
        """Return a list with the position of the members of a team.
        
        If an opponent is too far (> 5 square Manhattan distance),
        the corresponding entry will be None, as its position is unavailable.
        
        Keyword arguments:
        enemy_indices -- If True, returns the positions of the opposing team,
                    otherwise return the position of your own team.
                    (default: True)
        """
        indices = self.get_team_indices(game_state, enemy_indices=enemy_pos)
        positions = [game_state.getAgentPosition(idx) for idx in indices]
        return positions
    
    def get_team_distances(self, game_state, enemy_dist=True):
        """Return a list with the noisy distances from the members of a team.
        Distances will be returned as the real distance, +/- 6.
        
        Keyword arguments:
        enemy_dist -- If True, returns the distance of the opposing team,
                    otherwise return the distance of your own team.
                    (default: True)
        """
        indices = self.get_team_indices(game_state, enemy_indices=enemy_dist)
        distances = game_state.getAgentDistances()
        return [distances[idx] for idx in indices]

    def get_score(self, game_state):
        """Return your score.
        
        Your score is given by a number that is the difference between your
        score and the opponents score. This number is negative if you're losing.
        """
        if self.is_red:
            return game_state.getScore()
        else:
            return -game_state.getScore()

