"""Priority dictionary using binary heaps
    David Eppstein, UC Irvine, 8 Mar 2002
    
    Implements a data structure that acts almost 
    like a dictionary, with two modifications:
    (1) D.smallest() returns the value xvar minimizing D[xvar].  
          For this to work correctly, all values D[xvar] stored 
          in the dictionary must be comparable. 
    (2) iterating "for xvar in D" finds and removes the 
          items from D in sorted order.
          Each item is not removed until the next item is 
          requested, so D[xvar] will still
          return a useful value until the next iteration of the for-loop.
    Each operation takes logarithmic amortized time.
"""

from __future__ import generators


class priorityDictionary(dict):
    """"Modified dictionary that can return the minimum entry
    and iterating removes the items in sorted order.
    """
    
    def __init__(self):
        '''Initialize priorityDictionary by creating binary heap of 
        pairs (value,key).
        Note that changing or removing a dict entry will 
        not remove the old pair from the heap
        until it is found by smallest() or until the heap is rebuilt.
        '''
        self.__heap = []
        dict.__init__(self)

    def smallest(self):
        '''Find smallest item after removing deleted items from front of heap.
        '''
        if len(self) == 0:
            raise IndexError, "smallest of empty priorityDictionary"
        heap = self.__heap
        while heap[0][1] not in self or self[heap[0][1]] != heap[0][0]:
            last_item = heap.pop()
            insertion_point = 0
            while 1:
                small_child = 2 * insertion_point + 1
                if small_child + 1 < len(heap) and \
                    heap[small_child] > heap[small_child + 1]:
                    small_child += 1
                if small_child >= len(heap) or last_item <= heap[small_child]:
                    heap[insertion_point] = last_item
                    break
                heap[insertion_point] = heap[small_child]
                insertion_point = small_child
        return heap[0][1]
    
    def __iter__(self):
        '''Create destructive sorted iterator of priorityDictionary.'''
        def iterfn():
            """Internal function that is returned by __iter__"""
            while len(self) > 0:
                xvar = self.smallest()
                yield xvar
                del self[xvar]
        return iterfn()
    
    def __setitem__(self, key, val):
        '''Change value stored in dictionary and add corresponding 
        pair to heap. Rebuilds the heap if the number of deleted 
        items gets large, to avoid memory leakage.
        '''
        dict.__setitem__(self, key, val)
        heap = self.__heap
        if len(heap) > 2 * len(self):
            self.__heap = [(v, k) for k, v in self.iteritems()]
            self.__heap.sort()  # builtin sort faster(?) than O(n) heapify
        else:
            new_pair = (val, key)
            insertion_point = len(heap)
            heap.append(None)
            while insertion_point > 0 and \
                new_pair < heap[(insertion_point - 1) // 2]:
                heap[insertion_point] = heap[(insertion_point - 1) // 2]
                insertion_point = (insertion_point - 1) // 2
            heap[insertion_point] = new_pair

    def setdefault(self, key, val):
        '''Reimplement setdefault to pass through our customized __setitem__.'''
        if key not in self:
            self[key] = val
        return self[key]
