"""
KickAssAgents(tm)  --  Kicking your agents ass since 1878.

Simple Agents (KIS) that act on the basis of a combination of weight maps.
There are no states involved, it simply uses a weighted combination of
penalty maps for the factors food, enemies and partners.
"""

import numpy as np

from basic_agents import BasicAgentFactory, BasicAgent
from game import Directions


INFINITY = 1000000

NEIGHBOURS = [(-1,0), (1,0), (0,-1), (0,1)]

DIRECTIONS = {
    Directions.NORTH: (0,1),
    Directions.SOUTH: (0,-1),
    Directions.EAST: (1,0),
    Directions.WEST: (-1,0),
    Directions.STOP: (0,0),
    }

# algorithm parameters
# note that positive values are generally bad, e.g. distance to food
DEFAULT_HQ_PARAMS = {
    "ghost radius": 5,  # radius of effect for enemy ghosts
    "victim radius": 5,  # radius of effect for enemy pacmans
    
}

DEFAULT_AGENT_PARAMS = {
    "food multiplier": 1,
    "ghost multiplier": 2,  # multiplier for enemy ghosts
    "victim multiplier": -4,  # multiplier for enemy pacmans
    "partner radius": 6,  # radius of effect for partner
    "partner radius state factor": 1.0 / 2,  # radius depending on score
    "partner multiplier": 1,  # multiplier for partner
    "max stops": 2,  # maximum number of stop commands to avoid stalls
    }


class KickAssAgents(BasicAgentFactory):
    """Returns one offensive reflex agent and one defensive reflex agent"""
 
    def __init__(self, **args):
        """Initialize agent factory."""
        BasicAgentFactory.__init__(self, **args)
        self._agent_hq = KickAssHQ(DEFAULT_HQ_PARAMS)
        self.agents = []
 
    def getAgent(self, index):
        """Return a new agent."""
        # workaround for capture bug, it requests 3 agents
        if len(self.agents) >= 2:
            return BasicAgent(index)
        new_agent = KickAssAgent(index, self._agent_hq, DEFAULT_AGENT_PARAMS)
        self.agents.append(new_agent)
        return new_agent
    
    
class KickAssHQ(object):
    """This class coordinates the agents and stores the common data."""
    
    def __init__(self, params):
        """Initialize the internal variables.
        
        params -- Parameter dictionary, see example at top of this module.
        """
        self._params = params
        self.agents = []
        self.agents_acted = None  # dicts for my agents, True if acted in step
        # attributes that are initialized by initialize_hq
        self.initialized = False
        self.wall_map = None  # boolean array which is True for walls
        self._map_indices = None  # all index tuples for the map
        self._team = None  # 'red' or 'blue'
        # attributes which are updated in each step
        self.food_map = None
        self.ghost_map = None
        self.victim_map = None
        
    def initialize_hq(self, game_state):
        """Initialization with the initial state."""
        if game_state.isOnRedTeam(self.agents[0].index):
            self._team = 'red'
        else:
            self._team = 'blue'
        self.wall_map = np.array(game_state.getWalls().data)
        self._map_indices = [(i,j)
                        for i in range(self.wall_map.shape[0])
                        for j in range(self.wall_map.shape[1])]
        self.update_state(game_state)
        self.agents_acted = {}
        for agent in self.agents:
            self.agents_acted[agent] = True
        self.initialized = True
        
    def register_agent(self, agent):
        """Register an agent instance."""
        self.agents.append(agent)
        
    def update_state(self, game_state):
        """Is called once for each step by the agents."""
        self._update_maps(game_state)
    
    def _update_maps(self, game_state):
        """Update the internal maps."""
        # food map
        if self._team == 'red':
            food_position_map = np.array(game_state.getBlueFood().data)
        else:
            food_position_map = np.array(game_state.getRedFood().data)
        self.food_map = np.where(food_position_map, 0, -1)
        self.calc_distance_map(self.food_map)
        ## enemy maps
        # first assume team blue, which is on the left side
        if self._team == 'red':
            enemy_indices = game_state.getBlueTeamIndices()
        else:
            enemy_indices = game_state.getRedTeamIndices()
        enemy_positions = [game_state.getAgentPosition(index)
                            for index in enemy_indices] 
        ghost_positions = [pos for pos in enemy_positions
                           if pos and pos[0] < self.wall_map.shape[0] / 2]
        victim_positions = [pos for pos in enemy_positions
                            if pos and pos[0] > self.wall_map.shape[0] / 2]
        if self._team == 'red':
            ghost_positions, victim_positions = \
                                            victim_positions, ghost_positions
        # enemy agent in the middle
        ghost_positions += [pos for pos in enemy_positions
                            if pos and pos[0] == self.wall_map.shape[0] / 2]
        ## ghost map
        self.ghost_map = -1 * np.ones(self.wall_map.shape)
        for pos in ghost_positions:
            self.ghost_map[pos] = 0
        self.calc_distance_map(self.ghost_map,
                                max_dist=self._params["ghost radius"])
        # turn the distance around, the closer to the ghost the higher
        self.ghost_map = np.where(self.ghost_map == -1, 0,
                            -self.ghost_map + self._params["ghost radius"])
        ## victim map
        self.victim_map = -1 * np.ones(self.wall_map.shape)
        for pos in victim_positions:
            self.victim_map[pos] = 0
        self.calc_distance_map(self.victim_map,
                                max_dist=self._params["victim radius"])
        self.victim_map = np.where(self.victim_map == -1, 0,
                            -self.victim_map + self._params["victim radius"])
        
    def calc_distance_map(self, distance_map, max_dist=None):
        """Return a distance map.
        
        It takes into account the labyrint walls and does not overwrite the
        values at their positions.
        
        distance_map -- 2d array where the values are either distance 0 or
            negative. The negative entries will then be overwritten with the
            correct distance value.
        max_dist -- Integer that if provided (default is None) stops the
            distance calculation at the given value. The values at the
            unvisited positions are not modified.
        """
        updated_indices = [index for index in self._map_indices
                           if not distance_map[index] and
                              not self.wall_map[index]]
        unkown_indices = [index for index in self._map_indices
                          if distance_map[index] and not self.wall_map[index]]
        current_dist = 0
        while unkown_indices:
            current_dist += 1
            if max_dist and (current_dist > max_dist):
                return distance_map
            old_updated_indices = updated_indices
            updated_indices = []
            for index in old_updated_indices:
                for neigbour in NEIGHBOURS:
                    nb_index = (index[0] + neigbour[0], index[1] + neigbour[1])
                    if nb_index in unkown_indices:
                        distance_map[nb_index] = current_dist
                        updated_indices.append(nb_index)
                        unkown_indices.remove(nb_index)
        return distance_map
        
        
class KickAssAgent(BasicAgent):
    
    def __init__(self, index, agent_hq, params):
        """Initialize the agent.
        
        agent_hq -- KickAssHQ this agent belongs to.
        params -- Parameter dictionary, see example at top of this module.
        """
        super(KickAssAgent, self).__init__(index)
        self._agent_hq = agent_hq
        self._params = params
        self._agent_hq.register_agent(self)
        # counts the number of stop commands to prevent stalls
        self._stop_counter = 0
    
    def register_initial_state(self, game_state):
        super(KickAssAgent, self).register_initial_state(game_state)
        if not self._agent_hq.initialized:
            self._agent_hq.initialize_hq(game_state)
            
    def choose_action(self, game_state):
        """Update the maps and choose the action based on the combined map.
        
        The weights of the maps are currently constant.
        """
        ## check if this is the first call in the new step, update if yes
        if not (False in self._agent_hq.agents_acted.values()):
            self._agent_hq.update_state(game_state)
            for agent in self._agent_hq.agents:
                self._agent_hq.agents_acted[agent] = False
        self._agent_hq.agents_acted[self] = True
        ## get _team positions
        my_position = game_state.getAgentPosition(self.index)
        if self is self._agent_hq.agents[0]:
            partner = self._agent_hq.agents[1]
        else:
            partner = self._agent_hq.agents[0]
        partner_position = game_state.getAgentPosition(partner.index)
        ## combine the individual weight maps
        overall_map = self._agent_hq.food_map * self._params["food multiplier"]
        # add the walls
        overall_map[self._agent_hq.wall_map] = INFINITY
        # use enemy maps
        overall_map += (self._agent_hq.ghost_map *
                        self._params["ghost multiplier"])
        overall_map += (self._agent_hq.victim_map *
                        self._params["victim multiplier"])
        # the partner map has to be calculated here as it depends on the agent
        partner_radius = self._params["partner radius"]
        partner_radius_modifier = int(self.get_score(game_state) *
                              self._params["partner radius state factor"])
        partner_radius += partner_radius_modifier
        partner_map = -1 * np.ones(self._agent_hq.wall_map.shape)
        partner_map[partner_position] = 0
        self._agent_hq.calc_distance_map(partner_map, max_dist=partner_radius)
        partner_map = np.where(partner_map == -1, 0,
                               -partner_map + partner_radius)
        overall_map += (partner_map * self._params["partner multiplier"])
        ## now get the best movement option
        best_directions = []
        for direction, move in DIRECTIONS.items():
            index = (my_position[0] + move[0], my_position[1] + move[1])
            value = overall_map[index]
            best_directions.append((value, direction))
        best_directions.sort()
        best_directions = best_directions[::-1]
        best_direction = best_directions.pop()[1]
        if best_direction is Directions.STOP:
            self._stop_counter += 1
            if self._stop_counter > self._params["max stops"]:
                best_direction = best_directions.pop()[1]
                self._stop_counter = 0
        else:
            self._stop_counter = 0
        return best_direction
    
