from basic_agents import BasicAgentFactory, BasicAgent
from game import Directions
import capture
import captureAgents
import utils_pietro as pu
import random
import numpy
        
# ###################
# Agent factories
# ###################

class Goodfellas(BasicAgentFactory):
    def get_agents_list(self):
        return [ArnoldWinkelried, SmartPathFollower]

class PietroAgents(BasicAgentFactory):
    def get_agents_list(self):
        return [FoodyAgent, ArnoldWinkelried]

class SwissAgents(BasicAgentFactory):
    def get_agents_list(self):
        return [ArnoldWinkelried, WilhelmTell]

class DrunkAgentFactory(BasicAgentFactory):
    def get_agents_list(self):
        return [DrunkAgent, DrunkAgent]

# #########
# Agents
# #########

class DrunkAgent(BasicAgent):
    def choose_action(self, game_state):
        self.say(random.choice(['Hey!', 'Ho', 'Mrmmmf']))
        actions = game_state.getLegalActions(self.index)
        return random.choice(actions)
    
class SwissArmyKnifeAgent(BasicAgent):
    """
    Interesting internal variables:
    maze_graph -- a graph representation of the maze (a MazeGraph instance)
    maze_nodes -- a dictionary of maze coordinates to graph nodes
    maze[y][x] -- a boolean numpy array, True if there is a wall at (x,y)
    border_nodes -- a list of nodes on the boundary between allied and enemy
                    sides
    """
    
    def register_initial_state(self, game_state):
        super(SwissArmyKnifeAgent, self).register_initial_state(game_state)
        
        # create a graph representation of the maze
        self.maze = game_state.getWalls().data
        self.maze = numpy.array(self.maze, dtype='bool').T
        self.maze_graph, self.maze_nodes = pu.walls_to_graph(self.maze)
        
        # find points on border points between allied and enemy side
        print self.maze.shape, len(self.maze[0]) // 2
        half_x = len(self.maze[0]) // 2
        if self.is_red:
            half_x = half_x - 2
        else:
            half_x = half_x + 1
        notwall_y = numpy.logical_not(self.maze[:,half_x]).nonzero()
        self.border_nodes = [self.maze_nodes[(half_x, y)] for y in notwall_y[0]]
        
        # register start position (to determine whether you've been killed)
        self.start_position = game_state.getInitialAgentPosition(self.index)
        
    def is_killed(self, game_state):
        """Return True if the agent has been eaten on the previous turn."""
        return game_state.getAgentPosition(self.index) == self.start_position
    
    def get_food_coordinates(self, game_state, enemy_food=True):
        """Return the maze coordinates of food belonging to one of the teams.

        Keyword arguments:
        enemy_food -- If True, returns the food you're supposed to eat,
                    otherwise return the food you should protect.
                    (default: True)
        """
        food = self.get_food(game_state, enemy_food=enemy_food)
        food = numpy.array(food.data, 'bool')
        idx = food.nonzero()
        return zip(idx[0], idx[1])
    
    def is_on_our_side(self, x):
        """Return true if x is on allied side of the maze."""
        half = len(self.maze[0]) // 2
        if self.is_red:
            return x < half
        else:
            return x >= half
                    
    def get_closest_enemy_pos(self, game_state):
        """Returns the position of the closest visible enemy,
            None if none is available."""
        pos = game_state.getAgentPosition(self.index)
        enemy_pos = self.get_team_positions(game_state)
        enemy_pos = [epos for epos in enemy_pos if epos is not None]
        distances = [pu.manhattan(pos, epos) for epos in enemy_pos
                     if epos is not None]
        if len(distances) > 0:
            min_dist_idx = numpy.argmin(distances)
            # visible enemies could be visible to allied agents
            if distances[min_dist_idx] <= capture.SIGHT_RANGE:
                return enemy_pos[min_dist_idx]
        return None


class FoodyAgent(SwissArmyKnifeAgent):
    """This agent looks for the closest food and plans the shortest
    path to get, then follows the path.
    1) it does not try to avoid ghosts
    2) if you use two FoodyAgent, they'll try to do the same thing 
    """
    
    def register_initial_state(self, game_state):
        super(FoodyAgent, self).register_initial_state(game_state)
        self.has_plan = False
        
    def choose_action(self, game_state):
        pos = game_state.getAgentPosition(self.index)
        
        # check we have not been killed in the meanwhile
        if self.has_plan and self.is_killed(game_state):
            self.say("I'm so dead!")
            self.has_plan = False
            
        # plan path if currently idle
        if not self.has_plan:
            self.say("I'm hungry")
            # ### look for closest food position
            current_node = self.maze_nodes[pos]
            food = self.get_food_coordinates(game_state)
            food_nodes = [self.maze_nodes[pos] for pos in food]
            self.plan_nodes, self.plan_actions = \
                self.maze_graph.get_path_to_closest(current_node, food_nodes)
            if len(self.plan_actions) == 0:
                # it seems that sometimes there are 2 food in one spot
                # or something like that...
                return Directions.STOP
            self.has_plan = True
                    
        action = self.plan_actions.pop(0)
        # plan is finished
        if len(self.plan_actions) == 0:
            self.has_plan = False

        # print noisy distance from enemy
        who_am_I = 'ghost' if self.is_on_our_side(pos[0]) else 'PacMan'
        #self.say(str(self.get_team_distances(game_state, enemy_dist=True)))
        self.say("I'm a " + who_am_I)

        return action

class SmartPathFollower(SwissArmyKnifeAgent):
    """This agent looks for the nearest food, but is able to adapt to
    avoid nearby ghost and chase nearby pacmans."""
    def choose_action(self, game_state):
        pos = game_state.getAgentPosition(self.index)
        
        # create a new maze graph
        maze_graph, maze_nodes = self.maze_graph, self.maze_nodes
        maze_graph.cool(scale=0.7)

        # change weights if opponents are around
        opp_positions = self.get_team_positions(game_state)
        self.say(str(opp_positions))
        for opp_pos in opp_positions:
            if opp_pos is not None:
                if not self.is_on_our_side(opp_pos[0]):
                    maze_graph.update_weights_around(maze_nodes[opp_pos],
                                                     new_weights=100000., extent=4,
                                                     scale=0.1)
                elif len(self.maze[0]) // 2 - opp_pos[0] > 3:
                    maze_graph.update_weights_around(maze_nodes[opp_pos],
                                                     new_weights=1, extent=4)

        # stay away from allies
        allied_indices = self.get_team_indices(game_state, enemy_indices=False)
        allied_indices.remove(self.index)
        for idx in allied_indices:
            allied_pos = game_state.getAgentPosition(idx)
            maze_graph.update_weights_around(maze_nodes[allied_pos],
                                             new_weights=50., extent=3)
            

        # plan new path to closest food
        # ### look for closest food position
        current_node = maze_nodes[pos]
        food = self.get_food_coordinates(game_state)
        food_nodes = [maze_nodes[pos] for pos in food]
        self.plan_nodes, self.plan_actions = \
            maze_graph.get_weighted_path_to_closest(current_node, food_nodes)

        if len(self.plan_actions) == 0:
            # it seems that sometimes there are 2 food in one spot
            # or something like that...
            return Directions.STOP
        
        action = self.plan_actions.pop(0)

        # distance meter
        distances = self.get_team_distances(game_state)
        max_symbols = self.maze
        nr_symbols = min(distances) // 4

        return action
    

class ArnoldWinkelried(SwissArmyKnifeAgent):
    """This agent looks for a branch of the maze with a single entry
    and 3 or more pieces of food inside. It then goes there and stops
    right in front of it."""
    
    MOSTFOOD = 'MOSTFOOD'
    NEAREST = 'NEAREST'
    MODE = MOSTFOOD

    def register_initial_state(self, game_state):
        super(ArnoldWinkelried, self).register_initial_state(game_state)
        #CaptureAgent.registerInitialState(self, game_state)

        # form graph from walls
        self.has_plan = False
        # another copy of the graph to mistreat in the following
        maze_graph, maze_nodes = pu.walls_to_graph(self.maze)

        # look for dead ends on our side
        candidates = []
        def collect_dead_ends(node):
            x = node.data[0]
            if node.in_degree()==1 and self.is_on_our_side(x):
                candidates.append(node)
                
        maze_graph.bfs(maze_nodes[self.start_position],
                       visit_fct=collect_dead_ends)
        
        # for each dead end, trace back path up to the next crossing
        # and count food on the path
        our_food = self.get_food(game_state, enemy_food=False)
        consider = []
        for node in candidates:
            node.collapsed = []
            node.nfood = int(our_food[node.data[0]][node.data[1]])
            
            current = node
            while current.in_degree() <= 1:
                next = current.out_neighbors()[0]
                
                # remove current node and edges to next from graph
                maze_graph.remove_node(current)
                # update collapsed data
                next.collapsed = current.collapsed + [current]
                if not hasattr(next, 'nfood'):
                    next.nfood = int(our_food[next.data[0]][next.data[1]])
                next.nfood = next.nfood + current.nfood
                
                current = next
            
            # consider all nodes, given new rules    
            consider.append(current)

        # now that we have a list of branches with >= 3 foods,
        if self.MODE == self.MOSTFOOD:
            # choose the one with most food
            max_foodamount = max([node.nfood for node in consider])
            # get the closest node with the max amount of food
            goal_nodes = [self.maze_nodes[node.data] for node in consider
                          if node.nfood==max_foodamount]
        else:
            # choose the nearest with nfood >= 2
            goal_nodes = [self.maze_nodes[node.data] for node in consider
                          if node.nfood >= 2]
            
        # no appropriate spot found: panic
        # revert to DefensiveReflexAgent
        self.has_plan = len(goal_nodes) > 0
        if not self.has_plan:
            self.backup_agent = captureAgents.DefensiveReflexAgent(self.index)
            self.backup_agent.registerInitialState(game_state)
            return
       
        start_node = self.maze_nodes[self.start_position]
        self.plan_nodes, self.plan_actions = \
            maze_graph.get_path_to_closest(start_node, goal_nodes)
            
    def choose_action(self, game_state):
        if not self.has_plan:
            return self.backup_agent.chooseAction(game_state)
        if len(self.plan_actions) == 0:
            if self.is_red:
                team_txt = 'blue'
            else:
                team_txt = 'red'
            self.say('The %s team SUCKS!' % team_txt)
            return Directions.STOP
        else:
            action = self.plan_actions.pop(0)
            self.say(action)
            # plan is finished
            if len(self.plan_actions) == 0:
                self.goal_reached = True
            return action

# ============================
# Finite State Machine agent
# ============================

class AgentState(object):
    def __init__(self, agent):
        self.agent = agent
        
    def enter_state(self):
        pass
        
    def choose_action(self, game_state):
        return Directions.STOP

class FSMAgent(SwissArmyKnifeAgent):
    def __init__(self, index, start_state):
        super(FSMAgent, self).__init__(index)
        self.set_state(start_state)

    def set_state(self, state):
        state.enter_state()
        self.state = state
        
    def choose_action(self, game_state):
        return self.state.choose_action(game_state)
    
    def say(self, txt):
        #print self.index, txt
        super(FSMAgent, self).say(txt)
        
class WaitAtBorderState(AgentState):
    def choose_action(self, game_state):
        agent = self.agent
        agent.say('wait_border')
        closest_enemy_pos = agent.get_closest_enemy_pos(game_state)
        if closest_enemy_pos is None or agent.is_on_our_side(closest_enemy_pos[0]):
            agent.set_state(agent.goto_food_state)
            return agent.goto_food_state.choose_action(game_state)
        return Directions.STOP
    
class GoToBorderState(AgentState):
    def enter_state(self):
        self.has_plan = False
        
    def choose_action(self, game_state):
        agent = self.agent
        agent.say('goto_border')
        
        if not self.has_plan:
            pos = game_state.getAgentPosition(agent.index)
            start_node = agent.maze_nodes[pos]
            self.plan_nodes, self.plan_actions = \
                agent.maze_graph.get_path_to_closest(start_node, agent.border_nodes)
            self.has_plan = True
        
        action = self.plan_actions.pop(0)
        # plan is finished
        if len(self.plan_actions) == 0:
            self.has_plan = False
            agent.set_state(agent.wait_at_border_state)
            return agent.wait_at_border_state.choose_action(game_state)

        return action
    
class ChaseState(AgentState):
    def choose_action(self, game_state):
        agent = self.agent
        agent.say('chase')
        pos = game_state.getAgentPosition(agent.index)

        # get enemy to chase, if any
        closest_enemy_pos = agent.get_closest_enemy_pos(game_state)
        
        # enemy escaped or by taking
        if closest_enemy_pos is None:
            agent.set_state(agent.goto_food_state)
            return agent.goto_food_state.choose_action(game_state)

        current = agent.maze_nodes[pos]
        enemy = agent.maze_nodes[closest_enemy_pos]
        _, plan_actions = \
            agent.maze_graph.get_path_to_closest(current, [enemy])
        
        # enemy escaped or by taking this action we would become a pacman
        next_state = game_state.generateSuccessor(agent.index, plan_actions[0])
        next_pos = next_state.getAgentPosition(agent.index)
        if not agent.is_on_our_side(next_pos[0]):
            return Directions.STOP
            
        return plan_actions[0]

class GoToFoodState(AgentState):
    def enter_state(self):
        self.has_plan = False
        
    def choose_action(self, game_state):
        agent = self.agent
        agent.say('goto_food')
        pos = game_state.getAgentPosition(agent.index)

        closest_enemy_pos = agent.get_closest_enemy_pos(game_state)
        if closest_enemy_pos is not None:
            # start chase if enemy is close enough and we're ghosts
            if agent.is_on_our_side(pos[0]):
                agent.set_state(agent.chase_state)
                return agent.chase_state.choose_action(game_state)
            else:
                agent.set_state(agent.goto_border_state)
                return agent.goto_border_state.choose_action(game_state)
        
        # check we have not been killed in the meanwhile
        if self.has_plan and agent.is_killed(game_state):
            self.has_plan = False
            
        # plan path if currently idle
        if not self.has_plan:
            # ### look for closest food position
            current_node = agent.maze_nodes[pos]
            food = agent.get_food_coordinates(game_state)
            food_nodes = [agent.maze_nodes[pos] for pos in food]
            self.plan_nodes, self.plan_actions = \
                agent.maze_graph.get_path_to_closest(current_node, food_nodes)
            if len(self.plan_actions) == 0:
                # it seems that sometimes there are 2 food in one spot
                # or something like that...
                return Directions.STOP
            self.has_plan = True
                    
        action = self.plan_actions.pop(0)
        # plan is finished
        if len(self.plan_actions) == 0:
            self.has_plan = False

        return action
    
class WilhelmTell(FSMAgent):
    def __init__(self, index):
        # setup states
        self.goto_food_state = GoToFoodState(self)
        self.goto_border_state = GoToBorderState(self)
        self.wait_at_border_state = WaitAtBorderState(self)
        self.chase_state = ChaseState(self)
        self.stop_state = AgentState(self)
        super(WilhelmTell, self).__init__(index, self.goto_food_state)
        
    def register_initial_state(self, game_state):
        super(WilhelmTell, self).register_initial_state(game_state)
        