#!/usr/bin/env python
#coding=utf-8

"""
Some utils for some agents.
"""

from game import Directions
#from distanceCalculator import manhattanDistance
import util
import sys
import random
import numpy

import networkx as nx

#TODO: Graph stuff in other thread

class MazeAnalyser(object):
    """Maze analyser that uses networkx for easier analysis.
    
    (Maybe join with ShortestPath…)
    """
    def __init__(self, maze):
        self.maze = numpy.logical_not(numpy.array(maze, dtype='bool'))
        
        self.islands = []
        self.dead_ends = []
        self.streets = []
        self.t_crossings = []
        self.crossings = []
        
        self._path_length_cache = {}
        self.G = None

        # list of lists with x[0] dead end, x[-1] junction
        self.cul_de_sacs = {}
        
        self.init_positions()
        #print 'eval time for init: %.4f' % (time.time() - start)
        
        self.init_graph()
        
    def init_graph(self):
        G = nx.Graph()
        for pos, val in numpy.ndenumerate(self.maze):
            if not val: continue
            G.add_node(pos)
            points = [d + pos for d in CardinalDirections.values()]
            for p in points:
                if self.maze[p]:
                    G.add_edge(pos, p)
        self.G = G
        

    def init_positions(self):
        """Finds dead ends and returns a list of position tuples.
        """
        for position, value in numpy.ndenumerate(self.maze):
            if not value: continue
            # forget the border
            if 0 in position: continue
            if position[0] == self.maze.shape[0] - 1: continue
            if position[1] == self.maze.shape[1] - 1: continue
            # sum the 4 neighbouring points
            points = [d + position for d in CardinalDirections.values()]
            s = sum(self.maze[x] for x in points)
            if s == 0: self.islands.append(position)
            elif s == 1: self.dead_ends.append(position)
            elif s == 2: self.streets.append(position)
            elif s == 3: self.t_crossings.append(position)
            elif s == 4: self.crossings.append(position)

    def strategic_points(self, game_state, enemy=False):
        pass
    
    def before_dead_end(self, game_state, pos):
        neighbours = self.G[pos]
        dead_ends = []
        for n in neighbours:
            if len(self.G[n]) > 2: continue
            dead_ends += self.in_dead_end(game_state, n)
        return dead_ends
    
    def in_dead_end(self, game_state, pos):
        dead_ends = []
        neighbours = [pos]
        visited = []
        while neighbours:
            n = neighbours.pop()
            if n in visited: continue
            visited.append(n)
            new_neighbours = self.G[n]
            if len(new_neighbours) == 1:
                dead_ends.append(n)
            if len(new_neighbours) == 2:
                neighbours += new_neighbours
        return dead_ends


    def is_in_maze(self, pos):
        x, y = pos
        return self.maze[int(x)][int(y)]
    
    def get_shortest_path(self, pos_start, pos_end):
        """
        Returns the shortest path between two points.
        Should not be called if the Distancer is not ready yet.
        """
        path = []
        
        while pos_start != pos_end:
            d = self.get_direction_for_shortest_path(pos_start, pos_end)
            if not d:
                break
            path.append(d)
            new_direction = d[0]
            pos_start = CardinalDirections[new_direction] + pos_start
        return path
    
    
    def get_direction_for_shortest_path(self, pos_start, pos_end):
        # Make some noise, if necessary
        if not self.is_in_maze(pos_start):
            raise Exception("Start not in Maze")
        if not self.is_in_maze(pos_end):
            raise Exception("End not in Maze")
        
        #if not self.distancer.isReadyForMazeDistance:
        #    raise Exception("Not ready yet. Execution will give wrong results.")
        
        # Get all valid points in vicinity
        adjacent = []
        for direction in CardinalDirections:
            move = CardinalDirections[direction]
            pos_new = move + pos_start
            if pos_new.is_valid() and self.is_in_maze(pos_new):
                adjacent.append( (direction, pos_new) )
        
        closest_direction = []
        closest = sys.maxint
        # choose the best points
        for direction, pos_new in adjacent:
            
            dist_new = self.path_length(pos_new, pos_end)
            #dist_new = self.distancer.getDistance(pos_new, pos_end)
            
            if dist_new < closest:
                closest = dist_new
                closest_direction = [ (direction, pos_new) ]
            elif dist_new == closest:
                closest_direction.append( (direction, pos_new) )
            else:
                continue
        
        if closest_direction == []:
            return None
        # select one 
        return random.choice(closest_direction)
    
    def path_length(self, pos_start, pos_end, fear_factor = 0, avoid=[]):
        if not fear_factor:
            key = (pos_start, pos_end)
            if not key in self._path_length_cache:
                self._path_length_cache[key] = nx.dijkstra_path_length(self.G, pos_start, pos_end)
            return self._path_length_cache[key]
        else:
            enemy_G = self.G.copy()
            edges = enemy_G.edges(avoid)
            all_edges = edges
            all_edges += enemy_G.edges(edges)
            enemy_G.remove_edges_from(all_edges)
            for e in all_edges:
                enemy_G.add_edge( e[0], e[1], weight=fear_factor )
            return nx.dijkstra_path_length(enemy_G, pos_start, pos_end)
        
    #def path_length_with_enemy?    
    
class ShortestPath():
    """
    Calculates the shortest path between two points by stepping through it.
    """
    def __init__(self, maze_analyser):
        self.maze_analyser = maze_analyser
    
    def get_shortest_path_for_agent(self, agent, pos_end, game_state, fear_factor = 0, avoid=[]):
        """
        Returns the shortest path from the current position to pos_end. 
        If fear_factor > 0, enemies will be tracked. 
        """
        if fear_factor == 0:
            pos_start = game_state.getAgentPosition(agent.index)
            return self.get_shortest_path(pos_start, pos_end)
        else:
            pos_start = game_state.getAgentPosition(agent.index)
            return self.get_shortest_path(pos_start, pos_end, fear_factor, avoid)
        #values = [self.evaluate_distance(agent, game_state, a) for a in actions]
    
    def get_shortest_path(self, pos_start, pos_end, fear_factor=0, avoid=[]):
        """
        Returns the shortest path between two points.
        Should not be called if the Distancer is not ready yet.
        """
        path = []
        
        while pos_start != pos_end:
            d = self.get_direction_for_shortest_path(pos_start, pos_end, fear_factor=0, avoid=[])
            if not d:
                break
            path.append(d)
            new_direction = d[0]
            pos_start = CardinalDirections[new_direction] + pos_start
        return path
    
    
    def get_direction_for_shortest_path(self, pos_start, pos_end, fear_factor=0, avoid=[]):
        # Make some noise, if necessary
        if not self.maze_analyser.is_in_maze(pos_start):
            raise Exception("Start not in Maze")
        if not self.maze_analyser.is_in_maze(pos_end):
            raise Exception("End not in Maze")
                
        # Get all valid points in vicinity
        adjacent = []
        for direction in CardinalDirections:
            move = CardinalDirections[direction]
            pos_new = move + pos_start
            if pos_new.is_valid() and self.maze_analyser.is_in_maze(pos_new):
                adjacent.append( (direction, pos_new) )
        
        closest_direction = []
        closest = sys.maxint
        # choose the best points
        for direction, pos_new in adjacent:
            dist_new = self.maze_analyser.path_length(pos_new, pos_end, fear_factor=0, avoid=[])
            
            if dist_new < closest:
                closest = dist_new
                closest_direction = [ (direction, pos_new) ]
            elif dist_new == closest:
                closest_direction.append( (direction, pos_new) )
            else:
                continue
        
        if closest_direction == []:
            return None
        # select one 
        return random.choice(closest_direction)
        
    def length_of_path(self, path):
        l = 0
        for i in range(len(path)-1):
            l += len(self.get_shortest_path(path[i], path[i+1]))
        return l
    
    def solve_TSP(self, first_point, coordinates):
        coordinates.insert(0, first_point)

        size = len(coordinates)
        mst = numpy.zeros( (size, size), 'int')

        tsp_graph = nx.Graph()

        for i in range(size):
            for j in range(size):
                if j>=i: continue
                distance = self.maze_analyser.path_length(coordinates[i], coordinates[j])
                mst[i][j] = distance
                tsp_graph.add_edge( i, j, weight=distance )

        mst = nx.kruskal_mst(tsp_graph)
        edgelist = list(mst)
        #print sorted(edgelist)
        vertices = range(size)
        T=nx.Graph(edgelist)
        v = T.nodes()
        e = T.edges(data=True)
        #print e
        e = map(lambda x: (x[0], x[1], x[2]['weight']), e)
        set_e = set(e)
        old_set_e = set(e)
        for e in old_set_e:
            set_e.add( (e[1], e[0], e[2] ) )
        
        last_node = 0
        unvisited_nodes = vertices
        visited_edges = []
        visited_nodes = [last_node]
        child_edge_tree = {}

        unvisited_nodes.remove(last_node)
        while unvisited_nodes:
            # all child_edges
            child_edges = [ edges for edges in set_e if edges[0]==last_node  ]
            child_edge_tree[last_node] = child_edges
            
            #print child_edge_tree
            # get closest unvisited child_edge
            closest_dist = sys.maxint
            closest_child = None
            for child in child_edge_tree[last_node]:
                if child[1] in visited_nodes:
                    continue
                if child[2] < closest_dist:
                    closest_dist = child[2]
                    closest_child = child

            if closest_child is None:
                i = visited_nodes.index(last_node)
                last_node = visited_nodes[i-1]
            else:
                last_node = closest_child[1]
                visited_nodes += [last_node]
                unvisited_nodes.remove(last_node)

                visited_edges += [closest_child]                
            
        ordered_coords = []
        for n in visited_nodes[1:]:
            ordered_coords += [coordinates[n]]
        #print "done"
        return ordered_coords
        
    
    def solve_TSP_old(self, first_point, coordinates):
        """
        Tries to approximate an optimal order of points to visit. 
        """
        #print "TSP called: ", first_point, coordinates
        coordinates.insert(0, first_point)

        size = len(coordinates)
        mst = numpy.zeros( (size, size), 'int')

        edges = []
        for i in range(size):
            for j in range(size):
                if j>=i: continue
                distance = self.maze_analyser.path_length(coordinates[i], coordinates[j])
                mst[i][j] = distance
                edges.append( (i, j, distance) )
                #edges.append( (j, i, distance) )
                #mst[j][i] = distance

        vertices = range(size)
        v, e = prim(vertices, edges)
        
        set_e = set(e)
        old_set_e = set(e)
        for e in old_set_e:
            set_e.add( (e[1], e[0], e[2]) )
        
        last_node = 0
        unvisited_nodes = vertices
        visited_edges = []
        visited_nodes = [last_node]
        child_edge_tree = {}

        unvisited_nodes.remove(last_node)
        while unvisited_nodes:
            # all child_edges
            child_edges = [ edges for edges in set_e if edges[0]==last_node  ]
            child_edge_tree[last_node] = child_edges
            
            #print child_edge_tree
            # get closest unvisited child_edge
            closest_dist = sys.maxint
            closest_child = None
            for child in child_edge_tree[last_node]:
                if child[1] in visited_nodes:
                    continue
                if child[2] < closest_dist:
                    closest_dist = child[2]
                    closest_child = child

            if closest_child is None:
                i = visited_nodes.index(last_node)
                last_node = visited_nodes[i-1]
            else:
                last_node = closest_child[1]
                visited_nodes += [last_node]
                unvisited_nodes.remove(last_node)

                visited_edges += [closest_child]                
            
        ordered_coords = []
        for n in visited_nodes[1:]:
            ordered_coords += [coordinates[n]]
        #print "done"
        return ordered_coords    
        
    def get_successor(self, agent, game_state, action):
        """
        Finds the next successor which is a grid position (location tuple).
        """
        successor = game_state.generateSuccessor(agent.index, action)
        pos = successor.getAgentState(agent.index).getPosition()
        if pos != util.nearestPoint(pos): 
          # Only half a grid position was covered
            return successor.generateSuccessor(agent.index, action)
        else:
            return successor

class FlexiblePosition(tuple):
    """
    Small class that makes additions of position tuples just a little easier.
    """
    def __add__(self, other):
        res = []
        new = zip(self, other)
        for i in new:
            res.append(i[0]+i[1])
        return FlexiblePosition(res)
    
    def is_valid(self):
        for i in self:
            if i < 0: return False
        return True 

GO_NORTH = FlexiblePosition( (0, 1) )
GO_SOUTH = FlexiblePosition( (0, -1) )
GO_EAST = FlexiblePosition( (1, 0) )
GO_WEST = FlexiblePosition( (-1, 0) )
GO_NOT = FlexiblePosition( (0, 0) )

CardinalDirections = { Directions.NORTH: GO_NORTH,
                       Directions.EAST: GO_EAST,
                       Directions.SOUTH: GO_SOUTH,
                       Directions.WEST: GO_WEST }

CardinalDirections_Stop = { Directions.NORTH: GO_NORTH,
                       Directions.EAST: GO_EAST,
                       Directions.SOUTH: GO_SOUTH,
                       Directions.WEST: GO_WEST,
                       Directions.STOP: GO_NOT }

def neighbours(point, include_self=False):
    if include_self:
        return [n + point for n in CardinalDirections.values() + [GO_NOT]]
    else:
        return [n + point for n in CardinalDirections.values()]

def find_neighbour_of_in(point, some_ugly_iterable_mass, include_self=False):
    n = []
    for i in some_ugly_iterable_mass:
        if i in neighbours(point, include_self):
            n.append(i)
    return n

def kruskal(knoten, kanten):
    # knoten ist eine Liste von Knoten
    # kanten ist eine Liste von 3-Tupeln:
    #   (knoten1, knoten2, kosten)
    # Gibt ein Tupel ( knoten, kanten ) im selben
    # Format zurück
    t_knoten = [ knoten[0] ] 
    t_kanten = []
    zshg_komponenten = []
 
    sortierteKanten = sorted(kanten, key=lambda x: x[2])
    while len(t_knoten) != len(knoten):
        # Günstigste Kante nehmen
        kante = sortierteKanten.pop(0)
        # Die Kanten, die einen Kreis bilden würden,
        # entfernen. Dafür Zusammenhangskomponenten merken
        zshg_komp_a = [x for x in zshg_komponenten if kante[0] in x]
        zshg_komp_b = [x for x in zshg_komponenten if kante[1] in x]
        if zshg_komp_a and zshg_komp_b and zshg_komp_a == zshg_komp_b:
            continue
        elif zshg_komp_a and zshg_komp_b and zshg_komp_a != zshg_komp_b:
            zshg_komp_a += zshg_komp_b
            zshg_komponenten.remove(zshg_komp_b)
        elif zshg_komp_b:
            zshg_komp_b[0] += [ kante[0] ]
        elif zshg_komp_a:
            zshg_komp_a[0] += [ kante[1] ]
        else:
            zshg_komponenten += [ [ kante[0], kante[1] ] ]
        # Kante in Baum aufnehmen
        if kante[0] not in t_knoten: t_knoten += [ kante[0] ]
        if kante[1] not in t_knoten: t_knoten += [ kante[1] ]
        t_kanten += [ kante ]
 
        # Zu Lernzwecken:
        # print zshg_komponenten
 
    return (t_knoten, t_kanten)
 


def prim(vertices, edges):
    knoten = vertices
    kanten = edges 
    # From Wikibooks
    
    # knoten ist eine Liste von Knoten
    # kanten ist eine Liste von 3-Tupeln:
    #   (knoten1, knoten2, kosten)
    # Gibt ein Tupel ( knoten, kanten ) im selben
    # Format zurueck
 
    t_knoten = [ knoten[0] ]
    t_kanten = []
    while len(t_knoten) != len(knoten):
        akzeptable_kanten = [k for k in kanten if (k[0] in t_knoten) ^ (k[1] in t_knoten)]
        #akzeptable_kanten = filter(lambda x: (x[0] in t_knoten) ^ (x[1] in t_knoten), kanten)
        sortierte_kanten = sorted(akzeptable_kanten, key=lambda x: x[2])
        
        t_kanten += [ sortierte_kanten[0] ]
        t_knoten += [ sortierte_kanten[0][1 if sortierte_kanten[0][0] in t_knoten else 0] ]
    return t_knoten, t_kanten


class AgentTalk():
    def __init__(self):
        self._all_talk = {}
        self._current_text = []
        
        stripped = [ln.strip() for ln in self.AGENT_TALK.splitlines() if len(ln.strip())>1]
        all_talk = {}
        topic = ''
        for line in stripped:
            if line.startswith('%'):
                topic = line.lstrip('% ')
                if not all_talk.has_key(topic):
                    all_talk[topic] = []
                all_talk[topic].append([])
                index = 1
            elif line.startswith('-'):
                l = line.lstrip('- ')
                index = (index + 1) % 2
                text = (index, l)
                all_talk[topic][-1].append(text)
            else:
                text = (index, line)
                all_talk[topic][-1].append(text)

        self._all_talk = all_talk

    def select_talk(self, topic, index, other_index):
        """Selects a random talk from a given topic.
        index is starting, other_index is the dialogue partner
        """
        if not topic in self.topics:
            return False
        self._current_text = []
        for t in random.choice(self._all_talk[topic]):
            # doing it twice for longer duration
            if t[0] == 0:
                self._current_text.append((index, t[1]))
                self._current_text.append((index, t[1]))
            else:
                self._current_text.append((other_index, t[1]))
                self._current_text.append((other_index, t[1])) 
        
    def talk(self, index):
        """Returns the next line to be said by index.
        """
        if self._current_text:
            if self._current_text[0][0] == index:
                return self._current_text.pop(0)[1]
        
    def has_talk(self):
        """Returns true if a talk is currently selected.
        """
        if self._current_text: return True
        else: return False

    @property
    def topics(self):
        """A list of topics to be talked about.
        """
        return self._all_talk.keys()

    AGENT_TALK = u"""% Early
    - So, what do you think about the others?
    - They are horrible.
    % Talk
    - La vengeance se mange très-bien froide
    % Talk
    - If they move, kill ’em!
    % Help
    - Help me!
    - Yes, my friend.
    % Hunger
    - Hunger
    % Hunger
    - I’m so hungry
    % Hunger
    - Hunter needs food
    % Hunger
    - Foood!
    % Hunger
    - I could eat anything, man.
    % Despair
    - But I’ve got children, man.
    % Despair
    - We both don’t deserve this.
    % Defense
    - You know what?
    - Eh?
    - I think, I’m born a teacher.
    And I’m going to teach those schoolkids a lesson.
    % Kill
    - ‘K’ is for killing!
    % Kill
    - Keep your friends close,
    but your enemies closer.
    """