import unittest, numpy
import pietro_agents as pa
import utils_pietro as pu
import mdp
from game import Directions

class TestMazeGraph(unittest.TestCase):
    def setUp(self):
        walls = [[0,1,1,1,1],
                 [1,0,0,0,1],
                 [1,0,1,1,1],
                 [1,1,1,1,1]]
        self.walls = numpy.array(walls, dtype='b').tolist()
        self.graph, self.nodes_dict = pu.walls_to_graph(walls)
        
    def test_walls_to_graph(self):
        graph, nodes_dict = self.graph, self.nodes_dict
        self.assert_(len(graph.nodes)==5)
        self.assert_(len(graph.edges)==6)
        self.assertEqual(set(nodes_dict[(1,1)].out_neighbors()),
                         set([nodes_dict[(1,2)], nodes_dict[(2,1)]]))
        self.assertEqual(set(nodes_dict[(1,1)].in_neighbors()),
                         set([nodes_dict[(1,2)], nodes_dict[(2,1)]]))
        
    def test_planning(self):
        nodes = self.nodes_dict
        path, actions = self.graph.bfs_path(nodes[(1,2)], nodes[(3,1)])
        self.assertEqual(path,
                         [nodes[(1,2)], nodes[(1,1)], nodes[(2,1)], nodes[(3,1)]])
        self.assertEqual(actions, [Directions.SOUTH, Directions.EAST, Directions.EAST])
        self.assertRaises(mdp.graph.GraphException,
                          self.graph.bfs_path, nodes[(1,2)], nodes[(0,0)])
        
    def test_path_to_closest(self):
        nodes = self.nodes_dict
        path, actions = self.graph.get_path_to_closest(nodes[(1,1)],
                                                       [nodes[(1,2)], nodes[(3,1)]])
        self.assertEqual(actions, [Directions.NORTH])
        self.assertEqual(path, [nodes[(1,1)], nodes[(1,2)]])
    
class TestShortestPath(unittest.TestCase):
    
    def test_weighted_shortest_path(self):
        walls = [[1,1,1,1,1],
                 [1,0,0,0,1],
                 [1,0,1,0,1],
                 [1,0,0,0,1],
                 [1,1,1,1,1]]
        walls = numpy.array(walls, dtype='b').tolist()
        graph, nodes = pu.walls_to_graph(walls)
        edge = nodes[(1,1)].get_edges_out(to_=nodes[(2,1)])[0]
        edge.data[1] = 4.
        edge = nodes[(2,1)].get_edges_out(to_=nodes[(3,1)])[0]
        edge.data[1] = 4.
        
        path, actions = graph.weighted_shortest_path(nodes[(1,1)],
                                                     nodes[(3,1)])
        self.assertEqual(path, [nodes[(1,1)], nodes[(1,2)], nodes[(1,3)],
                                nodes[(2,3)], nodes[(3,3)], nodes[(3,2)],
                                nodes[(3,1)]])
        self.assertEqual(actions, [Directions.NORTH, Directions.NORTH,
                                   Directions.EAST, Directions.EAST,
                                   Directions.SOUTH, Directions.SOUTH])

    def test_closest_weighted_path(self):
        walls = [[1,1,1,1,1],
                 [1,0,0,0,1],
                 [1,0,1,0,1],
                 [1,0,0,0,1],
                 [1,1,1,1,1]]
        walls = numpy.array(walls, dtype='b').tolist()
        graph, nodes = pu.walls_to_graph(walls)
        edge = nodes[(1,1)].get_edges_out(to_=nodes[(2,1)])[0]
        edge.data[1] = 4.
        edge = nodes[(2,1)].get_edges_out(to_=nodes[(3,1)])[0]
        edge.data[1] = 4.

        path, actions = graph.get_weighted_path_to_closest(nodes[(1,1)],
                                                           [nodes[(3,2)],
                                                            nodes[(2,1)]])
        self.assertEqual(path, [nodes[(1,1)], nodes[(2,1)]])
        self.assertEqual(actions, [Directions.EAST])

        path, actions = graph.get_weighted_path_to_closest(nodes[(1,1)],
                                                           [nodes[(3,2)],
                                                            nodes[(3,1)]])
        self.assertEqual(path, [nodes[(1,1)], nodes[(1,2)], nodes[(1,3)],
                                nodes[(2,3)], nodes[(3,3)], nodes[(3,2)]])
        self.assertEqual(actions, [Directions.NORTH, Directions.NORTH,
                                   Directions.EAST, Directions.EAST,
                                   Directions.SOUTH])
        
    def test_update_weights(self):
        walls = [[1,1,0,1,1],
                 [1,1,0,1,1],
                 [0,0,0,0,0],
                 [1,1,0,1,1],
                 [1,1,0,1,1]]
        walls = numpy.array(walls, dtype='b').tolist()

        graph, nodes = pu.walls_to_graph(walls)
        graph.update_weights_around(nodes[(2,2)], new_weights=240.,
                                    extent=1, scale=0.1)
        edges = nodes[(2,2)].get_edges_out()
        for edge in edges:
            self.assertEqual(edge.data[1], 240.) 
        edge = nodes[(0,2)].get_edges_out()[0]
        self.assertEqual(edge.data[1], pu.DEFAULT_WEIGHT)
        
        graph, nodes = pu.walls_to_graph(walls)
        graph.update_weights_around(nodes[(2,2)], new_weights=240.,
                                    extent=2, scale=0.1)
        edges = nodes[(2,2)].get_edges_out()
        for edge in edges:
            self.assertEqual(edge.data[1], 240.) 
        edge = nodes[(2,0)].get_edges_out()[0]
        self.assertEqual(edge.data[1], 24.)
        
        
if __name__ == '__main__':
    unittest.main()