import mdp
import numpy
from mdp.graph import Graph
from game import Directions

INF = numpy.inf
# default weight for normal edges
DEFAULT_WEIGHT = 25.

def manhattan(pos1, pos2):
    return abs(pos1[0] - pos2[0]) + abs(pos1[1] - pos2[1])

# ####### Graphs and planning on graphs

class MazeGraph(Graph):
    """Extend mdp.graph.Graph with a method that returns a path
    between two nodes."""
    
    def traceback(self, root, goal, parents):
        """Using the results of a BFS, return a path between the root
        and the goal node.
        
        Inputs:
        root -- the root node
        goal -- the goal node
        parents -- the dictionary mapping sons to parents, returned by bfs_path
        
        Returns:
        path -- a list of nodes, from root to goal
        actions -- the data on the edges from root to goal (used to navigate
                   the maze)
        """
        # traceback path
        path = [goal]
        actions = []
        current = goal
        while current != root:
            # parents[current] = (parent, edge_to_parent)
            actions.insert(0, parents[current][1][0])
            current = parents[current][0]
            path.insert(0, current)
        return path, actions
            
    def bfs_path(self, root, goal=None):
        """Perform a Breadth First Search on a graph, starting at root.
        
        Inputs:
        root -- the root node
        goal -- (optional) goal node
        
        Outputs:
        If goal is not given:
        (list of nodes ordered by BFS, dictionary mapping nodes to their parents)
        If goal is given:
        (list of nodes on path from root to goal,
         data on the edges of the path, used to navigate the maze)
        """
        # result list containing the nodes in Breadth-First order
        bfs_list = []
        # keep track of all already visited nodes
        visited_nodes = { root: None }
        # parents of visited nodes
        parents = {}

        # queue (fifo) list
        bfs_queue = []
        bfs_queue.append(root)

        goal_found = False
        while len(bfs_queue) and not goal_found:
            # consider the next node in the queue
            node = bfs_queue.pop(0)
            bfs_list.append(node)
            # add all sons to the queue (if not already visited)
            for son_edge in node.get_edges_out():
                son = son_edge.get_tail()
                if son not in visited_nodes:
                    visited_nodes[son] = None
                    bfs_queue.append(son)
                    parents[son] = (node, son_edge.data)
                    if son == goal:
                        goal_found = True
                        break
        
        if goal is None:
            return bfs_list, parents
        
        if not goal_found:
            raise (mdp.graph.GraphException,
                   "Goal node was not reachable from root node")
        
        return self.traceback(root, goal, parents)
    
    def get_path_to_closest(self, start_node, goal_nodes):
        """Return a path from start_node to the closest of the nodes
        in goal_nodes.
        """
        bfs_nodes, parents = self.bfs_path(start_node)
        
        # ### look for closest goal
        # a lower index within the BFS results also corresponds to a
        # shorter distance from the current position
        goal_nodes_idx = [bfs_nodes.index(node) for node in goal_nodes]
        idx = numpy.argsort(goal_nodes_idx)
        closest = goal_nodes_idx[idx[0]]
        
        # ### get path to closest food
        plan_nodes, plan_actions = \
            self.traceback(start_node, bfs_nodes[closest], parents)

        return plan_nodes, plan_actions

    def weighted_shortest_path(self, root, goal=None):
        """Perform a weighted shortest path search on a graph, starting at root.
        
        Inputs:
        root -- the root node
        goal -- (optional) goal node
        
        Outputs:
        If goal is defined:
        (list of nodes on path from root to goal,
         data on the edges of the path, used to navigate the maze)
         If goal is None:
         (heap_queue, dictionary mapping nodes to their parents)
        """

        # remaining nodes to explore
        rest_nodes = set(self.nodes)

        # heap queue of all nodes
        bfs_queue = dict([(node, INF) for node in self.nodes])
        bfs_queue[root] = 0

        # parents of visited nodes
        parents = {}
        # edges connecting the nodes
        #edges = {}

        while len(rest_nodes):
            # consider the next node in the queue (the one with smallest distance)
            # find node with smallest distance

            _dist_list = [(bfs_queue[node], node) for node in rest_nodes]
            _dist_list.sort()
            node_dist, node = _dist_list[0]

            rest_nodes.remove(node)
            
            if node_dist == INF:
                # rest not reachable
                break

            # add all sons to the queue (if not already visited)
            for son_edge in node.get_edges_out():
                son = son_edge.get_tail()
                son_dist = node_dist + son_edge.data[1]
                if son_dist < bfs_queue.get(son, INF):
                    bfs_queue[son] = son_dist
                    parents[son] = (node, son_edge.data)
                    #edges[son] = son_edge
                    
        if goal is None:
            return bfs_queue, parents
        
        if goal not in parents.keys():
            raise (mdp.graph.GraphException,
                   "Goal node was not reachable from root node")

        return self.traceback(root, goal, parents)

    def get_weighted_path_to_closest(self, start_node, goal_nodes):
        """Return a path from start_node to the closest (as measured by weights)
        of the nodes in goal_nodes.
        """
        
        # perform a weighted search, then look at goal_nodes and get the closest one
        bfs_queue, parents = self.weighted_shortest_path(start_node)
        
        # ### look for closest goal
        min_dist = numpy.inf
        closest_node = None
        for node in goal_nodes:
            if bfs_queue[node] < min_dist:
                min_dist = bfs_queue[node]
                closest_node = node
        
        # ### get path to closest food
        plan_nodes, plan_actions = \
            self.traceback(start_node, closest_node, parents)

        return plan_nodes, plan_actions
    
    def update_weights_around(self, start_node, new_weights=DEFAULT_WEIGHT,
                              extent=0, scale=1.):
        """Starting at 'start_node' and down to 'extent' levels of neighbors,
        change the weights to 'new_weights*scale^neighbor_level'
        """
        visited = set()
        to_visit = [start_node]
        mult = 1.
        for _ in range(extent):
            next_visits = []
            for node in to_visit:
                visited.add(node)
                for neighbor in node.out_neighbors():
                    if neighbor in visited:
                        continue
                    # changes edges out
                    edges = node.get_edges_out(to_=neighbor)
                    for edge in edges:
                        edge.data[1] = new_weights
                    # changes edges in
                    edges = node.get_edges_in(from_=neighbor)
                    for edge in edges:
                        edge.data[1] = new_weights*scale
                    next_visits.append(neighbor)
            to_visit = next_visits
            mult *= scale

    def cool(self, scale=0.):
        for edge in self.edges:
            edge.data[1] = (edge.data[1] - DEFAULT_WEIGHT)*scale + DEFAULT_WEIGHT

def walls_to_graph(walls):
    """Transform a boolean 2D list (True = wall, False = free)
    in a graph. The data on the nodes correspond to their coordinates,
    data on edges is the actions to take to transition to that edge.
    Returns:
    graph -- a MazeGraph
    nodes_dict -- a dictionary that maps coordinates into nodes
                  e.g. nodes_dict[(3,4)] == node that represents coord x=3,y=4
    """
    width, height = len(walls[0]), len(walls)
    graph = MazeGraph()
    # define nodes for maze
    nodes_dict = {}
    for x in range(width):
        for y in range(height):
            if not walls[y][x]:
                nodes_dict[(x,y)] = graph.add_node((x,y))
    # add edges
    directions = [(-1,0,Directions.WEST),
                  (1,0,Directions.EAST),
                  (0,-1,Directions.SOUTH),
                  (0,1,Directions.NORTH)]
    for pos in nodes_dict:
        for dir in directions:
            neighbor = (pos[0] + dir[0], pos[1] + dir[1])
            if nodes_dict.has_key(neighbor):
                graph.add_edge(nodes_dict[pos], nodes_dict[neighbor],
                               data=[dir[2], DEFAULT_WEIGHT])
    return graph, nodes_dict

