#!/usr/bin/env python
#coding=utf-8

from game import Directions, Grid
from heapq import heappop, heappush
import numpy

WEST = Directions.WEST
EAST = Directions.EAST
SOUTH = Directions.SOUTH
NORTH = Directions.NORTH

class Node():
    ''' basic node class
        neighbours are stored as list of Node instances
        paths to other nodes as a dictionary mapping
        position tuples to a sorted list of path instances
    '''
    def __init__(self, pos):
        self.pos = pos
        self.neighbours = []
        self.paths = {}

    def __str__(self):
        desc = "Node:" + str(self.pos)
        if self.is_dead_end:
            desc += "DEAD_END!"
        desc += "\nNeighbours: "
        for n in self.neighbours:
            desc += "(" + str(n[0]) +", "
            desc +=  str(n[1].pos) + ") "
        desc += "\n" + str(self.paths) 
        return desc

    def add_neighbour(self, direction, Node):
        self.neighbours.append((direction, Node))

    @property
    def x(self):
        return self.pos[0]

    @property
    def y(self):
        return self.pos[1]

    @property
    def is_dead_end(self):
        return len(self.neighbours) is 1

class Path():
    ''' definition of a path through the pacman maze 
        contains start tuple, end tuple, and a list of actions
        on how to get there. Also some basic utilitys are provided
    '''
    def __init__(self, start, end, actions, positions):
        self.start = start
        self.end = end
        self.actions = actions
        self.positions = positions

    def __len__(self):
        return len(self.actions)

    def __cmp__(self, other):
        return cmp(len(self.actions), len(other.actions))

    def __str__(self):
        return "\n".join(["Start: "+ str(self.start), \
            "End: "+ str(self.end),  " ".join(self.actions)])

    def __eq__(self, other):
        return self.start == other.start and self.end == other.end and \
                len(self.actions) == len(other.actions)

    def pop(self):
        return self.actions.pop(0), self.positions.pop(0)

    def copy(path):
        return Path(path.start, path.end, list(path.actions))

    def backwards(self):
        self.start , self.end = self.end, self.start
        self.actions = [Directions.REVERSE[a] for a in self.actions]
        actions.reverse()
        positions.reverse()

    def gets_closer_to(self, positions, graph, dist):
        ''' check if this path moves closer to certain positions
                positions = list of tuples we want to move closer to
                dist = distance to be within
        '''
        if len(positions) is not 0:
            #dist_from_pos = [len(p) for p in graph.bfs_all(self.start, positions)]
            dist_from_pos = [3.0]
            for p1,p2 in [(p1,p2) for p1 in positions for p2 in  self.positions[:3]]:
                path = graph.bfs(p1,p2)
                if path is not None:
                    if len(path) < dist:
                        return True
class Graph():
    def __init__(self,walls):
        self.nodes = walls_to_list(walls)
        self.width = walls.width
        self.height = walls.height
        #self.fillin_dead_ends

    def __str__(self):
        return "\n".join([str(n) for n in self.nodes.values()])

    #def fillin_dead_ends(self):
    #    for node in self.nodes.values():
    #        if 
    #def fillin_dead_ens(self, node):
    #    dead_end_depth = 0
    #    seen = []
    #    to_visit = [node]
    #    current = None
    #    while len(to_visit) > 0:
    #        current = to_visit.pop(0)
    #            for (action, neighbour) in self.nodes[node].neighbours:

    def bfs_all(self, start, end_list, seen=None):
        paths = [self.bfs(start, end, seen=seen) for end in end_list]
        paths = [p for p in paths if p is not None]
        paths.sort()
        return paths

    def bfs_closest(self, start, end_list, seen=None):
        return self.bfs_all(start, end_list, seen=seen)[0]

    def bfs(self, start, end, seen=None):
        ''' breadth first search
            
            params:
                start - start position as tuple
                end   - end position as tuple
            returns:
                Path instance
        '''
        seen = set(seen) if seen is not None else set()
        to_visit = []
        heappush(to_visit,(0.0,start))
        parents = {start: None}
        if start in seen:
            seen.remove(start)
        while len(to_visit) is not 0:
            dist, node = heappop(to_visit)
            if node == end:
                break
            elif node in seen:
                continue
            elif type(node) is not tuple:
                continue
            else:
                for (action, neighbour) in self.nodes[node].neighbours:
                    if neighbour.pos not in seen:
                        push_goal = manhatten_distance(start,neighbour.pos),neighbour.pos
                        heappush(to_visit, push_goal)
                        parents[neighbour.pos] = node
                    seen.add(node)

        if end not in parents.keys():
            return None

        positions = []
        actions = []
        current = end
        while current != start:
            par = parents[current]
            for (action, neighbour) in self.nodes[par].neighbours:
                if neighbour.pos == current:
                    actions.append(action)
                    positions.append(current)
                    break
            current = par
        actions.reverse()
        positions.reverse()
        return Path(start, end, actions, positions)

    def node_closest_to(self, target):
        ''' get the closest node to the target '''
        l = [manhatten_distance(n, target) for n in self.nodes.keys()]
        return self.nodes.keys()[l.index(min(l))]

    def get_nodes_within_distance(self, center_nodes, dist):
        result = []
        # nodes is a list of position tuples
        for center in center_nodes:
            # first get all nodes that are within manhatten distance
            within = [n for n in self.nodes.keys() if
                    manhatten_distance(n,center) <= dist]
            # then for each of these, if its also within maze distance
            # add to the list of returns
            for w in within:
                path = self.bfs(center, w)
                if path is not None and len(path) <= dist:
                    result.append(w)
            result.append(center)
        return result


def walls_to_list(walls):
    ''' convert a 2-d boolean array to a dictionary mapping position tuples to
    Node instance '''
    nodes = {}
    for x,y in walls.asList(False):
        nodes[(x, y)] = Node((x, y))

    directions = [(-1,0,WEST), (1,0,EAST), (0,-1,SOUTH), (0,1,NORTH)]

    # each node knows about its neighbours and how to get there
    for (n,(dx,dy,action)) in [(n,d) for n in nodes.values() for d in directions]:
        neighbour = (n.x + dx, n.y + dy)
        if nodes.has_key(neighbour):
            n.add_neighbour( action, nodes[neighbour])

    return nodes

def manhatten_distance(pos1, pos2):
    return abs(pos1[0]-pos2[0])+ abs(pos1[1]-pos2[1])

def euclid_distance(pos1, pos2):
    #FIXME implement this
    return numpy.sqrt(abs(pos1[0]-pos2[0])**2 + abs(pos1[1]-pos2[1])**2)

def direction_tuple(pos1, pos2):
        ''' direction from pos1 to pos2 as tuple of game.Directions'''
        results = []
        xdiff = pos1[0] - pos2[0]
        ydiff = pos1[1] - pos2[1]
        if xdiff < 0:
            results.append(Directions.EAST)
        elif xdiff > 0:
            results.append(Directions.WEST)
        if ydiff < 0:
            results.append(Directions.NORTH)
        elif ydiff > 0:
            results.append(Directions.SOUTH)
        return results


if __name__ == '__main__':

    initial=[[1, 1, 1, 1, 1],
             [1, 0, 0, 0, 0],
             [1, 0, 1, 1, 0],
             [1, 0, 0, 0, 0],
             [1, 1, 1, 1, 1]]

    test_graph = Grid(width=2, height=2, initialValue=False)
    g = Graph(test_graph)
    print g

